"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Analyzer = exports.analyzer = exports.AnalyzedDocument = void 0;
const LSP = __importStar(require("vscode-languageserver"));
const vscode_languageserver_1 = require("vscode-languageserver");
const config_1 = require("./config");
const document_1 = require("./document");
const logger_1 = require("./logger");
const argparse_1 = require("./parsing/argparse");
const complete_1 = require("./parsing/complete");
const source_1 = require("./parsing/source");
const symbol_1 = require("./parsing/symbol");
const references_1 = require("./references");
const exec_1 = require("./utils/exec");
const file_operations_1 = require("./utils/file-operations");
const flatten_1 = require("./utils/flatten");
const node_types_1 = require("./utils/node-types");
const translation_1 = require("./utils/translation");
const tree_sitter_1 = require("./utils/tree-sitter");
const workspace_manager_1 = require("./utils/workspace-manager");
const validate_1 = require("./diagnostics/validate");
const barrel_1 = require("./parsing/barrel");
const parser_1 = require("./parser");
/**
 * Builder function to create an AnalyzedDocument object.
 */
var AnalyzedDocument;
(function (AnalyzedDocument) {
    function create(document, documentSymbols, tree, commandNodes = [], sourceNodes = []) {
        return {
            document,
            documentSymbols,
            tree,
            root: tree.rootNode,
            commandNodes,
            sourceNodes,
        };
    }
    AnalyzedDocument.create = create;
    function createFull(document, documentSymbols, tree) {
        const commandNodes = [];
        const sourceNodes = [];
        tree.rootNode.descendantsOfType('command').forEach(node => {
            if ((0, source_1.isSourceCommandWithArgument)(node))
                sourceNodes.push(node.child(1));
            commandNodes.push(node);
        });
        return {
            document,
            documentSymbols,
            tree,
            root: tree.rootNode,
            commandNodes,
            sourceNodes,
        };
    }
    AnalyzedDocument.createFull = createFull;
    function createPartial(document, tree) {
        const commandNodes = [];
        const sourceNodes = [];
        tree.rootNode.descendantsOfType('command').forEach(node => {
            if ((0, source_1.isSourceCommandWithArgument)(node))
                sourceNodes.push(node.child(1));
            commandNodes.push(node);
        });
        return {
            document,
            documentSymbols: [],
            tree,
            root: tree.rootNode,
            commandNodes,
            sourceNodes,
        };
    }
    AnalyzedDocument.createPartial = createPartial;
    function isPartial(analyzedDocument) {
        return analyzedDocument.documentSymbols.length === 0;
    }
    AnalyzedDocument.isPartial = isPartial;
    function isFull(analyzedDocument) {
        return analyzedDocument.documentSymbols.length > 0;
    }
    AnalyzedDocument.isFull = isFull;
})(AnalyzedDocument || (exports.AnalyzedDocument = AnalyzedDocument = {}));
/***
 * Handles analysis of documents and caching their symbols.
 *
 * Lots of server functionality is implemented here. Including, but not limited to:
 *   - tree sitter parsing
 *   - document analysis and caching
 *   - workspace/document symbol searching
 *   - background analysis performed on startup
 *
 * Requires a tree-sitter Parser instance to be initialized for usage.
 */
class Analyzer {
    parser;
    /**
     * The cached documents from all workspaces
     *   - keys are the document uris
     *   - values are the AnalyzedDocument objects
     */
    cache = new AnalyzedDocumentCache();
    /**
     * All of the global symbols throughout all workspaces in the server.
     * Methods that use this cache might try to limit symbols to a single workspace.
     *
     * The `globalSymbols.map` is a used to cache the symbols for quick access
     *   - keys are the symbol names
     *   - values are the FishSymbol objects
     */
    globalSymbols = new GlobalDefinitionCache();
    constructor(parser) {
        this.parser = parser;
    }
    /**
     * The method that is used to instantiate the singleton `analyzer`, to avoid
     * dependency injecting the analyzer in every utility that might need it.
     *
     * This method can be called during the `connection.onInitialize()` in the server,
     * or `beforeAll()` in a test-suite.
     *
     * It is okay to use the analyzer returned for testing purposes, however for
     * consistency throughout source code, please use the exported `analyzer` variable.
     */
    static async initialize() {
        const parser = await (0, parser_1.initializeParser)();
        exports.analyzer = new Analyzer(parser);
        return exports.analyzer;
    }
    /**
     * Perform full analysis on a LspDocument to build a AnalyzedDocument containing
     * useful information about the document. It will also add the information to both
     * the cache of AnalyzedDocuments and the global symbols cache.
     * @param document The LspDocument to analyze.
     * @returns An AnalyzedDocument object.
     */
    analyze(document) {
        const analyzedDocument = this.getAnalyzedDocument(document);
        this.cache.setDocument(document.uri, analyzedDocument);
        for (const symbol of (0, flatten_1.iterateNested)(...analyzedDocument.documentSymbols)) {
            if (symbol.isGlobal())
                this.globalSymbols.add(symbol);
        }
        return analyzedDocument;
    }
    /**
     * Take a path to a file and turns it into a LspDocument, to then be analyzed
     * and cached. This is useful for testing purposes, or for the rare occasion that
     * we need to analyze a file that is not yet a LspDocument.
     */
    analyzePath(rawFilePath) {
        const path = (0, translation_1.uriToPath)(rawFilePath);
        const document = file_operations_1.SyncFileHelper.loadDocumentSync(path);
        if (!document) {
            logger_1.logger.warning(`analyzer.analyzePath: ${path} not found`);
            return undefined;
        }
        return this.analyze(document);
    }
    /**
     * Use on documents where we can assume the document nodes aren't important.
     * This could mainly be summarized as any file in `$fish_complete_path/*.fish`
     * This greatly reduces the time it takes for huge workspaces to be analyzed,
     * by only retrieving the bare minimum of information required from completion
     * documents. Since completion documents are fully parsed, only once a request
     * is made that requires a completion document, we are able to avoid building their
     * document symbols here. Conversely, this means that if we were to use this method
     * instead of the
     */
    analyzePartial(document) {
        const tree = this.parser.parse(document.getText());
        const analyzedDocument = AnalyzedDocument.createPartial(document, tree);
        this.cache.setDocument(document.uri, analyzedDocument);
        return analyzedDocument;
    }
    /**
     * Helper method to get the AnalyzedDocument.
     * Retrieves the parsed AST from tree-sitter's parser, processes the DocumentSymbols,
     * stores the commands used in the document, and collects all the sourced command
     * SyntaxNode's that might contain a sourced file.
     * @param LspDocument The LspDocument to analyze.
     * @returns An AnalyzedDocument object.
     */
    getAnalyzedDocument(document) {
        const tree = this.parser.parse(document.getText());
        const documentSymbols = (0, symbol_1.processNestedTree)(document, tree.rootNode);
        return AnalyzedDocument.createFull(document, documentSymbols, tree);
    }
    /**
     * Analyze a workspace and all its documents.
     * Documents that are already analyzed will be skipped.
     * For documents that are autoloaded completions, we
     */
    async analyzeWorkspace(workspace, progress = undefined, callbackfn = (text) => logger_1.logger.log(`analyzer.analyzerWorkspace(${workspace.name})`, text)) {
        const startTime = performance.now();
        if (workspace.isAnalyzed()) {
            callbackfn(`[fish-lsp] workspace ${workspace.name} already analyzed`);
            progress?.done();
            return { count: 0, workspace, duration: '0.00' };
        }
        // progress?.begin(workspace.name, 0, 'Analyzing workspace', true);
        const docs = workspace.pendingDocuments();
        const maxSize = Math.min(docs.length, config_1.config.fish_lsp_max_background_files);
        const currentDocuments = workspace.pendingDocuments().slice(0, maxSize);
        // Helper function to delay execution
        const delay = (ms) => new Promise(resolve => setTimeout(resolve, ms));
        // Calculate adaptive delay and batch size based on document count
        const BATCH_SIZE = Math.max(1, Math.floor(currentDocuments.length / 20));
        const UPDATE_DELAY = currentDocuments.length > 100 ? 10 : 25; // Shorter delay for large sets
        let lastUpdateTime = 0;
        const MIN_UPDATE_INTERVAL = 15; // Minimum ms between visual updates
        currentDocuments.forEach(async (doc, idx) => {
            try {
                if (doc.getAutoloadType() === 'completions') {
                    this.analyzePartial(doc);
                }
                else {
                    this.analyze(doc);
                }
                workspace.uris.markIndexed(doc.uri);
                const reportPercent = Math.ceil(idx / maxSize * 100);
                progress?.report(reportPercent, `Analyzing ${idx}/${docs.length} files`);
            }
            catch (err) {
                logger_1.logger.log(`[fish-lsp] ERROR analyzing workspace '${workspace.name}' (${err?.toString() || ''})`);
            }
            const currentTime = performance.now();
            const isLastItem = idx === currentDocuments.length - 1;
            const isBatchEnd = idx % BATCH_SIZE === BATCH_SIZE - 1;
            const timeToUpdate = currentTime - lastUpdateTime > MIN_UPDATE_INTERVAL;
            if (isLastItem || isBatchEnd && timeToUpdate) {
                const percentage = Math.ceil((idx + 1) / maxSize * 100);
                progress?.report(`${percentage}% Analyzing ${idx + 1}/${maxSize} ${maxSize > 1 ? 'documents' : 'document'}`);
                lastUpdateTime = currentTime;
                // Add a small delay for visual perception
                await delay(UPDATE_DELAY);
            }
        });
        progress?.done();
        const endTime = performance.now();
        const duration = ((endTime - startTime) / 1000).toFixed(2); // Convert to seconds with 2 decimal places
        const count = currentDocuments.length;
        const message = `Analyzed ${count} document${count > 1 ? 's' : ''} in ${duration}s`;
        callbackfn(message);
        return {
            count: currentDocuments.length,
            workspace: workspace,
            duration,
        };
    }
    /**
     * Update the configuration for the current workspace based on the symbols found in the document
     *
     * This is used to update the configuration for the current workspace when a user changes
     * a fish-lsp environment variable in the workspace.
     */
    updateConfigInWorkspace(documentUri) {
        const workspace = workspace_manager_1.workspaceManager.current;
        let symbols = this.getFlatDocumentSymbols(documentUri).filter(symbol => symbol.kind === vscode_languageserver_1.SymbolKind.Variable && Object.keys(config_1.config).includes(symbol.name));
        if (!workspace || !config_1.config.fish_lsp_single_workspace_support) {
            if (symbols.length === 0) {
                const prev = config_1.config.fish_lsp_single_workspace_support;
                Object.assign(config_1.config, (0, config_1.getDefaultConfiguration)());
                config_1.config.fish_lsp_single_workspace_support = prev;
                return;
            }
            (0, config_1.updateBasedOnSymbols)(symbols);
            return;
        }
        symbols = this.findSymbols((sym, doc) => {
            if (doc && workspace.contains(doc?.uri))
                return false;
            if (sym.kind === vscode_languageserver_1.SymbolKind.Variable && Object.keys(config_1.config).includes(sym.name)) {
                return true;
            }
            return false;
        });
        if (symbols.length > 0) {
            (0, config_1.updateBasedOnSymbols)(symbols);
        }
    }
    /**
     * Return the first FishSymbol seen that could be defined by the given position.
     */
    findDocumentSymbol(document, position) {
        const symbols = (0, flatten_1.flattenNested)(...this.cache.getDocumentSymbols(document.uri));
        return symbols.find((symbol) => {
            return (0, tree_sitter_1.isPositionWithinRange)(position, symbol.selectionRange);
        });
    }
    /**
     * Return all FishSymbols seen that could be defined by the given position.
     */
    findDocumentSymbols(document, position) {
        const symbols = (0, flatten_1.flattenNested)(...this.cache.getDocumentSymbols(document.uri));
        return symbols.filter((symbol) => {
            return (0, tree_sitter_1.isPositionWithinRange)(position, symbol.selectionRange);
        });
    }
    /**
     * Search through all the documents in the cache, and return the first symbol found
     * that matches the callback function.
     */
    findSymbol(callbackfn) {
        const currentWs = workspace_manager_1.workspaceManager.current;
        const uris = this.cache.uris().filter(uri => currentWs ? currentWs?.contains(uri) : true);
        for (const uri of uris) {
            const symbols = this.cache.getFlatDocumentSymbols(uri);
            const document = this.cache.getDocument(uri)?.document;
            const symbol = symbols.find(s => callbackfn(s, document));
            if (symbol) {
                return symbol;
            }
        }
        return undefined;
    }
    /**
     * Search through all the documents in the cache, and return all symbols found
     */
    findSymbols(callbackfn) {
        const currentWs = workspace_manager_1.workspaceManager.current;
        const uris = this.cache.uris().filter(uri => currentWs ? currentWs?.contains(uri) : true);
        const symbols = [];
        for (const uri of uris) {
            const document = this.cache.getDocument(uri)?.document;
            const symbols = this.getFlatDocumentSymbols(document.uri);
            const newSymbols = symbols.filter(s => callbackfn(s, document));
            if (newSymbols) {
                symbols.push(...newSymbols);
            }
        }
        return symbols;
    }
    /**
     * Search through all the documents in the cache, and return the first node found
     */
    findNode(callbackfn) {
        const uris = this.cache.uris();
        for (const uri of uris) {
            const root = this.cache.getRootNode(uri);
            const document = this.cache.getDocument(uri).document;
            if (!root || !document)
                continue;
            const node = (0, tree_sitter_1.getChildNodes)(root).find((n) => callbackfn(n, document));
            if (node) {
                return node;
            }
        }
        return undefined;
    }
    /**
     * Search through all the documents in the cache, and return all nodes found (with their uris)
     */
    findNodes(callbackfn) {
        const currentWs = workspace_manager_1.workspaceManager.current;
        const uris = this.cache.uris().filter(uri => currentWs ? currentWs?.contains(uri) : uri);
        const result = [];
        for (const uri of uris) {
            const root = this.cache.getRootNode(uri);
            const document = this.cache.getDocument(uri).document;
            if (!root || !document)
                continue;
            const nodes = (0, tree_sitter_1.getChildNodes)(root).filter((node) => callbackfn(node, document));
            if (nodes.length > 0) {
                result.push({ uri: document.uri, nodes });
            }
        }
        return result;
    }
    /**
     * Collect all the global symbols in the workspace, and the document symbols usable
     * at the requests position. DocumentSymbols that are not in the position's scope are
     * excluded from the result array of FishSymbols.
     *
     * This method is mostly notably used for providing the symbols in
     * `server.onCompletion()` requests.
     *
     * @param document The LspDocument to search in
     * @param position The position to search at
     * @returns {FishSymbol[]} A flat array of FishSymbols that are usable at the given position
     */
    allSymbolsAccessibleAtPosition(document, position) {
        // Set to avoid duplicate symbols
        const symbolNames = new Set();
        // add the local symbols
        const symbols = (0, flatten_1.flattenNested)(...this.cache.getDocumentSymbols(document.uri))
            .filter((symbol) => symbol.scope.containsPosition(position));
        symbols.forEach((symbol) => symbolNames.add(symbol.name));
        // add the sourced symbols
        const sourcedUris = this.collectReachableSources(document.uri, position);
        for (const sourcedUri of Array.from(sourcedUris)) {
            const sourcedSymbols = this.cache.getFlatDocumentSymbols(sourcedUri)
                .filter(s => !symbolNames.has(s.name)
                && (0, node_types_1.isTopLevelDefinition)(s.focusedNode)
                && s.uri !== document.uri);
            symbols.push(...sourcedSymbols);
            sourcedSymbols.forEach((symbol) => symbolNames.add(symbol.name));
        }
        // add the global symbols
        for (const globalSymbol of this.globalSymbols.allSymbols) {
            // skip any symbols that are already in the result so that
            // next conditionals don't have to consider duplicate symbols
            if (symbolNames.has(globalSymbol.name))
                continue;
            // any global symbol not in the document
            if (globalSymbol.uri !== document.uri) {
                symbols.push(globalSymbol);
                symbolNames.add(globalSymbol.name);
                // any symbol in the document that is globally scoped
            }
            else if (globalSymbol.uri === document.uri) {
                symbols.push(globalSymbol);
                symbolNames.add(globalSymbol.name);
            }
        }
        return symbols;
    }
    /**
     * method that returns all the workspaceSymbols that are in the same scope as the given
     * shell
     * @returns {WorkspaceSymbol[]} array of all symbols
     */
    getWorkspaceSymbols(query = '') {
        const workspace = workspace_manager_1.workspaceManager.current;
        logger_1.logger.log({ searching: workspace?.path, query });
        return this.globalSymbols.allSymbols
            .filter(symbol => workspace?.contains(symbol.uri) || symbol.uri === workspace?.uri)
            .map((s) => s.toWorkspaceSymbol())
            .filter((symbol) => {
            return symbol.name.startsWith(query);
        });
    }
    /**
     * Utility function to get the definitions of a symbol at a given position.
     */
    getDefinitionHelper(document, position) {
        const symbols = [];
        const localSymbols = this.getFlatDocumentSymbols(document.uri);
        const toFind = this.wordAtPoint(document.uri, position.line, position.character);
        const nodeToFind = this.nodeAtPoint(document.uri, position.line, position.character);
        if (!toFind || !nodeToFind)
            return [];
        const localSymbol = localSymbols.find((s) => {
            return s.name === toFind && (0, tree_sitter_1.containsRange)(s.selectionRange, (0, tree_sitter_1.getRange)(nodeToFind));
        });
        if (localSymbol) {
            symbols.push(localSymbol);
        }
        else {
            const toAdd = localSymbols.filter((s) => {
                const variableBefore = s.kind === vscode_languageserver_1.SymbolKind.Variable ? (0, tree_sitter_1.precedesRange)(s.selectionRange, (0, tree_sitter_1.getRange)(nodeToFind)) : true;
                return (s.name === toFind
                    && (0, tree_sitter_1.containsRange)((0, tree_sitter_1.getRange)(s.scope.scopeNode), (0, tree_sitter_1.getRange)(nodeToFind))
                    && variableBefore);
            });
            symbols.push(...toAdd);
        }
        if (!symbols.length) {
            symbols.push(...this.globalSymbols.find(toFind));
        }
        return symbols;
    }
    /**
     * Get the first definition of a position that we can find.
     * Will first
     */
    getDefinition(document, position) {
        const symbols = this.getDefinitionHelper(document, position);
        const word = this.wordAtPoint(document.uri, position.line, position.character);
        const node = this.nodeAtPoint(document.uri, position.line, position.character);
        const startTime = performance.now();
        if (node && (0, barrel_1.isExportVariableDefinitionName)(node)) {
            return symbols.find(s => s.name === word) || symbols.pop();
        }
        if (node && (0, node_types_1.isAliasDefinitionName)(node)) {
            return symbols.find(s => s.name === word) || symbols.pop();
        }
        if (node && (0, argparse_1.isArgparseVariableDefinitionName)(node)) {
            return this.getFlatDocumentSymbols(document.uri).findLast(s => s.containsPosition(position)) || symbols.pop();
        }
        if (node && (0, complete_1.isCompletionSymbol)(node)) {
            logger_1.logger.debug({
                isCompletionSymbol: true,
            });
            const completionSymbols = this.getFlatCompletionSymbols(document.uri);
            const completionSymbol = completionSymbols.find(s => s.equalsNode(node));
            if (!completionSymbol) {
                return null;
            }
            const symbol = this.findSymbol((s) => completionSymbol.equalsArgparse(s));
            const endTime = performance.now();
            const duration = ((endTime - startTime) / 1000).toFixed(2); // Convert to seconds with 2 decimal places
            logger_1.logger.debug({
                isCompletionSymbol: true,
                duration: `${duration} ms`,
                symbol: symbol?.name,
            });
            if (symbol) {
                return symbol;
            }
        }
        if (node && (0, node_types_1.isOption)(node)) {
            const symbol = this.findSymbol((s) => {
                if (s.parent && s.fishKind === 'ARGPARSE') {
                    return node.parent?.firstNamedChild?.text === s.parent?.name &&
                        s.parent.isGlobal() &&
                        node.text.startsWith(s.argparseFlag);
                }
                return false;
            });
            const endTIme = performance.now();
            logger_1.logger.debug({
                isOption: true,
                node: node.text,
                parent: node?.parent?.text,
                symbol: symbol?.name || '',
                duration: `${endTIme - startTime} ms`,
            });
            if (symbol)
                return symbol;
        }
        return symbols.pop() || null;
    }
    /**
     * Get all the definition locations of a position that we can find
     */
    getDefinitionLocation(document, position) {
        // handle source argument definition location
        const node = this.nodeAtPoint(document.uri, position.line, position.character);
        // check that the node (or its parent) is a `source` command argument
        if (node && (0, source_1.isSourceCommandArgumentName)(node)) {
            logger_1.logger.log({
                isSourceCommandArgumentName: node.text,
                node: true,
                parent: false,
            });
            return this.getSourceDefinitionLocation(node);
        }
        if (node && node.parent && (0, source_1.isSourceCommandArgumentName)(node.parent)) {
            logger_1.logger.log({
                isSourceCommandArgumentName: node.parent.text,
                node: false,
                parent: true,
            });
            return this.getSourceDefinitionLocation(node.parent);
        }
        // check if we have a symbol defined at the position
        const symbol = this.getDefinition(document, position);
        if (symbol)
            return [vscode_languageserver_1.Location.create(symbol.uri, symbol.selectionRange)];
        // This is currently the only location where `config.fish_lsp_single_workspace_support` is used.
        // It allows users to go-to-definition on commands that are not in the current workspace.
        if (!config_1.config.fish_lsp_single_workspace_support && workspace_manager_1.workspaceManager.current) {
            const node = this.nodeAtPoint(document.uri, position.line, position.character);
            if (node && (0, node_types_1.isCommandName)(node)) {
                const text = node.text.toString();
                const locations = (0, exec_1.execCommandLocations)(text);
                for (const { uri, path } of locations) {
                    const content = file_operations_1.SyncFileHelper.read(path, 'utf8');
                    const doc = document_1.LspDocument.createTextDocumentItem(uri, content);
                    workspace_manager_1.workspaceManager.handleOpenDocument(doc);
                }
                return locations.map(({ uri }) => vscode_languageserver_1.Location.create(uri, {
                    start: { line: 0, character: 0 },
                    end: { line: 0, character: 0 },
                }));
            }
        }
        return [];
    }
    /**
     * Here we can allow the user to use completion locations for the implementation.
     */
    getImplementation(document, position) {
        const definition = this.getDefinition(document, position);
        if (!definition)
            return [];
        const locations = (0, references_1.implementationLocation)(this, document, position);
        return locations;
    }
    /**
     * Gets the location of the sourced file for the given source command argument name node.
     */
    getSourceDefinitionLocation(node) {
        if (node && (0, source_1.isSourceCommandArgumentName)(node)) {
            const expanded = (0, source_1.getExpandedSourcedFilenameNode)(node);
            let sourceDoc = this.getDocumentFromPath(expanded);
            if (!sourceDoc) {
                this.analyzePath(expanded); // find the filepath & analyze it
                sourceDoc = this.getDocumentFromPath(expanded); // reset the sourceDoc to new value
            }
            if (sourceDoc) {
                return [
                    vscode_languageserver_1.Location.create(sourceDoc.uri, LSP.Range.create(0, 0, 0, 0)),
                ];
            }
        }
        return [];
    }
    /**
     * Get the hover from the given position in the document, if it exists.
     * This is either a symbol, a manpage, or a fish-shell shipped function.
     * Other hovers are shown are shown if this method can't find any (defined in `./hover.ts`).
     */
    getHover(document, position) {
        const tree = this.getTree(document.uri);
        const node = this.nodeAtPoint(document.uri, position.line, position.character);
        if (!tree || !node)
            return null;
        const symbol = this.getDefinition(document, position) ||
            this.globalSymbols.findFirst(node.text);
        if (!symbol)
            return null;
        logger_1.logger.log(`analyzer.getHover: ${symbol.name}`, {
            name: symbol.name,
            uri: symbol.uri,
            detail: symbol.detail,
            text: symbol.node.text,
            kind: (0, translation_1.symbolKindToString)(symbol.kind),
        });
        return symbol.toHover();
    }
    /**
     * Returns the tree-sitter tree for the given documentUri.
     * If the document is not in the cache, it will cache it and return the tree.
     *
     * @NOTE: we use `documentUri` here instead of LspDocument's because it simplifies
     *        testing and is more consistently available in the server.
     *
     * @param documentUri - the uri of the document to get the tree for
     * @return {Tree | undefined} - the tree for the document, or undefined if the document is not in the cache
     */
    getTree(documentUri) {
        if (this.cache.hasUri(documentUri)) {
            return this.cache.getDocument(documentUri)?.tree;
        }
        return this.analyzePath((0, translation_1.uriToPath)(documentUri))?.tree;
    }
    /**
     * Finds the rootnode given a LspDocument. If useCache is set to false, it will
     * use the parser to parse the document passed in, and then return the rootNode.
     */
    getRootNode(documentUri) {
        return this.cache.getParsedTree(documentUri)?.rootNode;
    }
    /**
     * Returns the document from the cache. If the document is not in the cache,
     * it will return undefined.
     */
    getDocument(documentUri) {
        return this.cache.getDocument(documentUri)?.document;
    }
    /**
     * Returns the document from the cache if the document is in the cache.
     */
    getDocumentFromPath(path) {
        const uri = (0, translation_1.pathToUri)(path);
        return this.getDocument(uri);
    }
    /**
     * Returns the FishSymbol[] array in the cache for the given documentUri.
     * The result is a nested array (tree) of FishSymbol[] items
     */
    getDocumentSymbols(documentUri) {
        return this.cache.getDocumentSymbols(documentUri);
    }
    /**
     * Returns the flat array of FishSymbol[] for the given documentUri.
     * Iterating through the result will allow you to reach every symbol in the documentUri.
     */
    getFlatDocumentSymbols(documentUri) {
        return this.cache.getFlatDocumentSymbols(documentUri);
    }
    /**
     * Returns a list of symbols similar to a DocumentSymbol array, but
     * instead of using that data type, we use our custom CompletionSymbol to define completions
     *
     * NOTE: while this method's visibility is public, it is really more of a utility
     *       for the `getGlobalArgparseLocations()` function in `src/parsing/argparse.ts`
     *
     * @param documentUri - the uri of the document to get the completions for
     * @returns {CompletionSymbol[]} - an array of CompletionSymbol objects
     */
    getFlatCompletionSymbols(documentUri) {
        const doc = this.cache.getDocument(documentUri);
        if (!doc)
            return [];
        const { document, commandNodes } = doc;
        // get the `complete` SyntaxNode[]
        const childrenSymbols = commandNodes.filter(n => (0, complete_1.isCompletionCommandDefinition)(n));
        // build the CompletionSymbol[] for the entire document
        const result = [];
        for (const child of childrenSymbols) {
            result.push(...(0, complete_1.processCompletion)(document, child));
        }
        return result;
    }
    /**
     * Returns a list of all the nodes in the document.
     */
    getNodes(documentUri) {
        const document = this.cache.getDocument(documentUri)?.document;
        if (!document) {
            return [];
        }
        return (0, tree_sitter_1.getChildNodes)(this.parser.parse(document.getText()).rootNode);
    }
    /**
     * Returns a list of all the diagnostics in the document (if the document is analyzed)
     * @param documentUri - the uri of the document to get the diagnostics for
     * @returns {Diagnostic[]} - an array of Diagnostic objects
     */
    getDiagnostics(documentUri) {
        const doc = this.getDocument(documentUri);
        const root = this.getRootNode(documentUri);
        if (!doc || !root) {
            return [];
        }
        return (0, validate_1.getDiagnostics)(root, doc);
    }
    /**
     * Utility to collect all the sources in the input documentUri, or if specified
     * it will only collect the included sources from the sources parameter
     * @param documentUri - the uri of the document to collect sources from
     * @param sources - the sources to collect from (optional set to narrow results)
     * @returns {Set<string>} - a flat set of all the sourceUri's reachable from the input sources
     */
    collectSources(documentUri, sources = this.cache.getSources(documentUri)) {
        const visited = new Set();
        const collectionStack = Array.from(sources);
        while (collectionStack.length > 0) {
            const source = collectionStack.pop();
            if (visited.has(source))
                continue;
            visited.add(source);
            if (file_operations_1.SyncFileHelper.isDirectory((0, translation_1.uriToPath)(source)))
                continue;
            if (!file_operations_1.SyncFileHelper.isFile((0, translation_1.uriToPath)(source)))
                continue;
            const cahedSourceDoc = this.cache.hasUri(source)
                ? this.cache.getDocument(source)
                : this.analyzePath((0, translation_1.uriToPath)(source));
            if (!cahedSourceDoc)
                continue;
            const sourced = this.cache.getSources(cahedSourceDoc.document.uri);
            collectionStack.push(...Array.from(sourced));
        }
        return visited;
    }
    /**
     * Collects all the sourceUri's that are reachable from the given documentUri at Position
     * @param documentUri - the uri of the document to collect sources from
     * @param position - the position to collect sources from
     * @returns {Set<string>} - a set of all the sourceUri's in the document before the position
     */
    collectReachableSources(documentUri, position) {
        const currentNode = this.nodeAtPoint(documentUri, position.line, position.character);
        let currentParent;
        if (currentNode)
            currentParent = (0, node_types_1.findParentFunction)(currentNode);
        const sourceNodes = this.cache.getSourceNodes(documentUri)
            .filter(node => {
            if ((0, node_types_1.isTopLevelDefinition)(node) && (0, tree_sitter_1.isPositionAfter)((0, tree_sitter_1.getRange)(node).start, position)) {
                return true;
            }
            const parentFunction = (0, node_types_1.findParentFunction)(node);
            if (currentParent && parentFunction?.equals(currentParent) && (0, tree_sitter_1.isPositionAfter)((0, tree_sitter_1.getRange)(node).start, position)) {
                return true;
            }
            return false;
        });
        const sources = new Set();
        for (const node of sourceNodes) {
            const sourced = (0, source_1.getExpandedSourcedFilenameNode)(node);
            if (sourced) {
                sources.add((0, translation_1.pathToUri)(sourced));
            }
        }
        return this.collectSources(documentUri, sources);
    }
    /**
     * Collects all the sourceUri's that are in the documentUri
     * @param documentUri - the uri of the document to collect sources from
     * @returns {Set<string>} - a set of all the sourceUri's in the document
     */
    collectAllSources(documentUri) {
        const allSources = this.collectSources(documentUri);
        for (const source of Array.from(allSources)) {
            const sourceDoc = this.cache.getDocument(source);
            if (!sourceDoc) {
                this.analyzePath(source);
            }
        }
        return allSources;
    }
    /**
     * Returns an object to be deconstructed, for the onComplete function in the server.
     * This function is necessary because the normal onComplete parse of the LspDocument
     * will commonly throw errors (user is incomplete typing a command, etc.). To avoid
     * inaccurate parses for the entire document, we instead parse just the current line
     * that the user is on, and send it to the shell script to complete.
     *
     * @Note: the position should not edited (pass in the direct position from the CompletionParams)
     *
     * @returns
     *        line - the string output of the line the cursor is on
     *        lineRootNode - the rootNode for the line that the cursor is on
     *        lineCurrentNode - the last node in the line
     */
    parseCurrentLine(document, position) {
        const line = document
            .getLineBeforeCursor(position)
            .replace(/^(.*)\n$/, '$1') || '';
        const word = this.wordAtPoint(document.uri, position.line, Math.max(position.character - 1, 0)) || '';
        const lineRootNode = this.parser.parse(line).rootNode;
        const lineLastNode = lineRootNode.descendantForPosition({
            row: 0,
            column: line.length - 1,
        });
        return { line, word, lineRootNode, lineLastNode };
    }
    wordAtPoint(uri, line, column) {
        const node = this.nodeAtPoint(uri, line, column);
        if (!node || node.childCount > 0 || node.text.trim() === '') {
            return null;
        }
        // check if the current word is a node that contains a `=` sign, therefore
        // we don't want to return the whole word, but only the part before the `=`
        if ((0, node_types_1.isAliasDefinitionName)(node) ||
            (0, barrel_1.isExportVariableDefinitionName)(node))
            return node.text.split('=')[0].trim();
        return node.text.trim();
    }
    /**
     * Find the node at the given point.
     */
    nodeAtPoint(uri, line, column) {
        const tree = this.cache.getParsedTree(uri);
        if (!tree?.rootNode) {
            // Check for lacking rootNode (due to failed parse?)
            return null;
        }
        return tree.rootNode.descendantForPosition({ row: line, column });
    }
    /**
     * Find the name of the command at the given point.
     */
    commandNameAtPoint(uri, line, column) {
        let node = this.nodeAtPoint(uri, line, column);
        while (node && !(0, node_types_1.isCommand)(node)) {
            node = node.parent;
        }
        if (!node)
            return null;
        const firstChild = node.firstNamedChild;
        if (!firstChild || !(0, node_types_1.isCommandName)(firstChild))
            return null;
        return firstChild.text.trim();
    }
    /**
     * Get the text at the given location, using the range of the location to find the text
     * inside the range.
     * Super helpful for debugging Locations like references, renames, definitions, etc.
     */
    getTextAtLocation(location) {
        const document = this.cache.getDocument(location.uri);
        if (!document) {
            return '';
        }
        const text = document.document.getText(location.range);
        return text;
    }
}
exports.Analyzer = Analyzer;
/**
 * The cache for all of the analyzer's global FishSymbol's across all workspaces
 * analyzed.
 *
 * The enternal map uses the name of the symbol as the key, and the value is an array
 * of FishSymbol's that have the same name. This is because a symbol can be defined
 * multiple times in different scopes/workspaces, and we want to keep track of all of them.
 */
class GlobalDefinitionCache {
    _definitions;
    constructor(_definitions = new Map()) {
        this._definitions = _definitions;
    }
    add(symbol) {
        const current = this._definitions.get(symbol.name) || [];
        if (!current.some(s => s.equals(symbol))) {
            current.push(symbol);
        }
        this._definitions.set(symbol.name, current);
    }
    find(name) {
        return this._definitions.get(name) || [];
    }
    findFirst(name) {
        const symbols = this.find(name);
        if (symbols.length === 0) {
            return undefined;
        }
        return symbols[0];
    }
    has(name) {
        return this._definitions.has(name);
    }
    uniqueSymbols() {
        const unique = [];
        this.allNames.forEach(name => {
            const u = this.findFirst(name);
            if (u) {
                unique.push(u);
            }
        });
        return unique;
    }
    get allSymbols() {
        const all = [];
        for (const [_, symbols] of this._definitions.entries()) {
            all.push(...symbols);
        }
        return all;
    }
    get allNames() {
        return [...this._definitions.keys()];
    }
    get map() {
        return this._definitions;
    }
}
/**
 * The cache for all of the analyzed documents in the server.
 *
 * The internal map uses the uri of the document as the key, and the value is
 * the AnalyzedDocument object that contains:
 *   - LspDocument
 *   - FishSymbols (the definitions in the Document)
 *   - tree (from tree-sitter)
 *   - `source` command arguments, SyntaxNode[]
 *   - commands used in the document (array of strings)
 */
class AnalyzedDocumentCache {
    _documents;
    constructor(_documents = new Map()) {
        this._documents = _documents;
    }
    uris() {
        return [...this._documents.keys()];
    }
    setDocument(uri, analyzedDocument) {
        this._documents.set(uri, analyzedDocument);
    }
    getDocument(uri) {
        if (!this._documents.has(uri)) {
            return undefined;
        }
        return this._documents.get(uri);
    }
    hasUri(uri) {
        return this._documents.has(uri);
    }
    updateUri(oldUri, newUri) {
        const oldValue = this.getDocument(oldUri);
        if (oldValue) {
            this._documents.delete(oldUri);
            this._documents.set(newUri, oldValue);
        }
    }
    getDocumentSymbols(uri) {
        return this._documents.get(uri)?.documentSymbols || [];
    }
    getFlatDocumentSymbols(uri) {
        return (0, flatten_1.flattenNested)(...this.getDocumentSymbols(uri));
    }
    getCommands(uri) {
        return this._documents.get(uri)?.commandNodes || [];
    }
    getRootNode(uri) {
        return this.getParsedTree(uri)?.rootNode;
    }
    getParsedTree(uri) {
        return this._documents.get(uri)?.tree;
    }
    getSymbolTree(uri) {
        const document = this.getDocument(uri);
        if (!document) {
            return [];
        }
        return document.documentSymbols;
    }
    getSources(uri) {
        const document = this.getDocument(uri);
        if (!document) {
            return new Set();
        }
        const result = new Set();
        const sourceNodes = document.sourceNodes.map(node => (0, source_1.getExpandedSourcedFilenameNode)(node)).filter(s => !!s);
        for (const source of sourceNodes) {
            const sourceUri = (0, translation_1.pathToUri)(source);
            result.add(sourceUri);
        }
        return result;
    }
    getSourceNodes(uri) {
        const document = this.getDocument(uri);
        if (!document) {
            return [];
        }
        return document.sourceNodes;
    }
    clear(uri) {
        this._documents.delete(uri);
    }
}
