"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DiagnosticCommentsHandler = exports.DIAGNOSTIC_COMMENT_REGEX = void 0;
exports.isDiagnosticComment = isDiagnosticComment;
exports.isValidErrorCode = isValidErrorCode;
exports.parseDiagnosticComment = parseDiagnosticComment;
const node_types_1 = require("../utils/node-types");
const error_codes_1 = require("./error-codes");
const config_1 = require("../config");
/**
 * Regular expression to match fish-lsp diagnostic control comments
 * Matches patterns like:
 * # @fish-lsp-disable
 * # @fish-lsp-enable
 * # @fish-lsp-disable 1001 1002
 * # @fish-lsp-enable 1001
 * # @fish-lsp-disable-next-line
 * # @fish-lsp-disable-next-line 3002 3001
 */
exports.DIAGNOSTIC_COMMENT_REGEX = /^#\s*@fish-lsp-(disable|enable)(?:-(next-line))?\s*([0-9\s]*)?$/;
/**
 * Checks if a node is a diagnostic control comment
 * @param node The syntax node to check
 * @returns true if the node is a diagnostic control comment
 */
function isDiagnosticComment(node) {
    if (!(0, node_types_1.isComment)(node))
        return false;
    return exports.DIAGNOSTIC_COMMENT_REGEX.test(node.text.trim());
}
function isValidErrorCode(code) {
    return Object.values(error_codes_1.ErrorCodes).includes(code);
}
/**
 * Parses a diagnostic comment node into its components
 * @param node The syntax node to parse
 * @returns DiagnosticComment object containing the parsed information
 */
function parseDiagnosticComment(node) {
    if (!isDiagnosticComment(node))
        return null;
    const match = node.text.trim().match(exports.DIAGNOSTIC_COMMENT_REGEX);
    if (!match)
        return null;
    const [, action, nextLine, codesStr] = match;
    const codeStrings = codesStr ? codesStr.trim().split(/\s+/) : [];
    // Parse the diagnostic codes if present
    const parsedCodes = codeStrings
        .map(codeStr => parseInt(codeStr, 10))
        .filter(code => !isNaN(code));
    const validCodes = [];
    const invalidCodes = [];
    codeStrings.forEach((codeStr, idx) => {
        const code = parsedCodes[idx];
        if (code && !isNaN(code) && isValidErrorCode(code)) {
            validCodes.push(code);
        }
        else {
            invalidCodes.push(codeStr);
        }
    });
    return {
        action: action,
        target: nextLine ? 'next-line' : 'line',
        codes: validCodes.length > 0 ? validCodes : error_codes_1.ErrorCodes.allErrorCodes,
        lineNumber: node.startPosition.row,
        invalidCodes: invalidCodes.length > 0 ? invalidCodes : undefined,
    };
}
class DiagnosticCommentsHandler {
    // Original stack-based state for compatibility during parsing
    stateStack = [];
    // Track all control points (sorted by line number) for position-based lookups
    controlPoints = [];
    // Map of line numbers to their effective states (calculated at the end)
    lineStateMap = new Map();
    // Track invalid codes for reporting
    invalidCodeWarnings = new Map();
    // Cached enabled comments for current state
    enabledComments = [];
    constructor() {
        // Initialize with global state
        this.pushState(this.initialState);
        this.enabledComments = Array.from(this.currentState.enabledCodes);
    }
    get initialState() {
        return {
            enabledCodes: new Set(this.globalEnabledCodes()),
            comment: {
                action: 'enable',
                target: 'line',
                codes: this.globalEnabledCodes(),
                lineNumber: -1,
            },
        };
    }
    get currentState() {
        return this.stateStack[this.stateStack.length - 1];
    }
    globalEnabledCodes() {
        if (config_1.config.fish_lsp_diagnostic_disable_error_codes.length > 0) {
            return error_codes_1.ErrorCodes.allErrorCodes.filter(code => !config_1.config.fish_lsp_diagnostic_disable_error_codes.includes(code)).filter(code => error_codes_1.ErrorCodes.nonDeprecatedErrorCodes.some(e => e.code === code));
        }
        return error_codes_1.ErrorCodes.allErrorCodes.filter(code => error_codes_1.ErrorCodes.nonDeprecatedErrorCodes.some(e => e.code === code));
    }
    pushState(state) {
        this.stateStack.push(state);
    }
    popState() {
        if (this.stateStack.length > 1) { // Keep at least the global state
            this.stateStack.pop();
            this.enabledComments = Array.from(this.currentState.enabledCodes);
        }
    }
    /**
     * Process a node for diagnostic comments
     * This maintains both the stack state and records control points
     */
    handleNode(node) {
        // Clean up any expired next-line comments
        this.cleanupNextLineComments(node.startPosition.row);
        // Early return if not a diagnostic comment
        if (!isDiagnosticComment(node)) {
            return;
        }
        const comment = parseDiagnosticComment(node);
        if (!comment)
            return;
        // Track invalid codes if present
        if (comment.invalidCodes && comment.invalidCodes.length > 0) {
            this.invalidCodeWarnings.set(comment.lineNumber, comment.invalidCodes);
        }
        // Process the comment for both backward compatibility and position-based lookups
        this.processComment(comment);
    }
    processComment(comment) {
        // Update stack-based state (for backward compatibility)
        const newEnabledCodes = new Set(this.currentState.enabledCodes);
        if (comment.action === 'disable') {
            comment.codes.forEach(code => newEnabledCodes.delete(code));
        }
        else {
            comment.codes.forEach(code => newEnabledCodes.add(code));
        }
        const newState = {
            enabledCodes: newEnabledCodes,
            comment,
            invalidCodes: comment.invalidCodes,
        };
        // Update control points for position-based lookups
        const controlPoint = {
            line: comment.lineNumber,
            action: comment.action,
            codes: comment.codes,
            isNextLine: comment.target === 'next-line',
        };
        this.controlPoints.push(controlPoint);
        // Keep control points sorted by line number
        this.controlPoints.sort((a, b) => a.line - b.line);
        if (comment.target === 'next-line') {
            // For next-line, we'll push a new state that will be popped after the line
            this.pushState(newState);
        }
        else {
            // For regular comments, we'll replace the current state
            if (this.stateStack.length > 1) {
                this.popState(); // Remove the current state
            }
            this.pushState(newState);
        }
        this.enabledComments = Array.from(newEnabledCodes);
    }
    cleanupNextLineComments(currentLine) {
        while (this.stateStack.length > 1 && // Keep global state
            this.currentState.comment.target === 'next-line' &&
            currentLine > this.currentState.comment.lineNumber + 1) {
            this.popState();
        }
    }
    /**
     * This method is called when all nodes have been processed
     * It computes the effective state for each line in the document
     */
    finalizeStateMap(maxLine) {
        // Start with initial state
        let currentState = {
            enabledCodes: new Set(this.globalEnabledCodes()),
        };
        // Process all regular control points first
        const regularPoints = this.controlPoints.filter(p => !p.isNextLine);
        const nextLinePoints = new Map();
        // Group next-line control points by target line
        for (const point of this.controlPoints) {
            if (point.isNextLine) {
                const targetLine = point.line + 1;
                const existing = nextLinePoints.get(targetLine) || [];
                existing.push({ ...point, line: targetLine });
                nextLinePoints.set(targetLine, existing);
            }
        }
        // Build line by line state
        for (let line = 0; line <= maxLine; line++) {
            // Apply regular control points for this line
            for (const point of regularPoints) {
                if (point.line <= line) {
                    this.applyControlPointToState(currentState, point);
                }
            }
            // Save the state before applying next-line directives
            const baseState = {
                enabledCodes: new Set(currentState.enabledCodes),
            };
            // Apply next-line directives for this line only
            const nextLineDirs = nextLinePoints.get(line) || [];
            for (const directive of nextLineDirs) {
                this.applyControlPointToState(currentState, directive);
            }
            // Store state for this line
            this.lineStateMap.set(line, {
                enabledCodes: new Set(currentState.enabledCodes),
            });
            // Restore base state after next-line directives
            if (nextLineDirs.length > 0) {
                currentState = baseState;
            }
        }
    }
    applyControlPointToState(state, point) {
        if (point.action === 'disable') {
            // Disable specified codes
            for (const code of point.codes) {
                state.enabledCodes.delete(code);
            }
        }
        else {
            // Enable specified codes
            for (const code of point.codes) {
                state.enabledCodes.add(code);
            }
        }
    }
    isCodeEnabledAtNode(code, node) {
        const position = { line: node.startPosition.row, character: node.startPosition.column };
        return this.isCodeEnabledAtPosition(code, position);
    }
    /**
     * Check if a specific diagnostic code is enabled at a given position
     * Will use the pre-computed state if available, otherwise computes on-demand
     */
    isCodeEnabledAtPosition(code, position) {
        if (this.lineStateMap.has(position.line)) {
            // Use pre-computed state if available
            const state = this.lineStateMap.get(position.line);
            return state.enabledCodes.has(code);
        }
        // Compute state on-demand if not pre-computed
        return this.computeStateAtPosition(position).enabledCodes.has(code);
    }
    /**
     * Compute state at a position on-demand (used if finalizeStateMap hasn't been called)
     */
    computeStateAtPosition(position) {
        // Start with global state
        const state = {
            enabledCodes: new Set(this.globalEnabledCodes()),
        };
        // Apply all regular control points up to this position
        for (const point of this.controlPoints) {
            if (point.line > position.line) {
                break; // Skip control points after this position
            }
            if (!point.isNextLine && point.line <= position.line) {
                this.applyControlPointToState(state, point);
            }
            // Apply next-line directives for the specific line
            if (point.isNextLine && point.line + 1 === position.line) {
                this.applyControlPointToState(state, { ...point, line: position.line });
            }
        }
        return state;
    }
    /**
     * Check if a specific diagnostic code is enabled in the current state
     * This is for backward compatibility during parsing
     */
    isCodeEnabled(code) {
        return this.currentState.enabledCodes.has(code);
    }
    getStackDepth() {
        return this.stateStack.length;
    }
    getCurrentState() {
        return this.currentState;
    }
    /**
     * For debugging/testing - get verbose state information
     */
    getCurrentStateVerbose() {
        const currentState = this.getCurrentState();
        const disabledCodes = error_codes_1.ErrorCodes.allErrorCodes.filter(e => !currentState.enabledCodes.has(e));
        const enabledCodes = Array.from(currentState.enabledCodes)
            .map(e => error_codes_1.ErrorCodes.codes[e].code)
            .concat(disabledCodes)
            .sort((a, b) => a - b)
            .map(item => {
            if (disabledCodes.includes(item))
                return '....';
            return item;
        })
            .join(' | ');
        const invalidCodes = Array.from(this.invalidCodeWarnings.entries())
            .map(([line, codes]) => `${line}: ${codes.join(' | ')}`);
        const lineStates = Array.from(this.lineStateMap.entries())
            .map(([line, state]) => `Line ${line}: ${Array.from(state.enabledCodes).length} enabled codes`);
        return {
            depth: this.getStackDepth(),
            enabledCodes: enabledCodes,
            invalidCodes: invalidCodes,
            currentState: {
                action: currentState.comment.action,
                target: currentState.comment.target,
                codes: currentState.comment?.codes.join(' | '),
                lineNumber: currentState.comment.lineNumber,
            },
            controlPoints: this.controlPoints.length,
            lineStates: lineStates.slice(0, 10), // Show first 10 for brevity
        };
    }
}
exports.DiagnosticCommentsHandler = DiagnosticCommentsHandler;
