"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.isPossibleDiagnosticComment = isPossibleDiagnosticComment;
exports.detectInvalidDiagnosticCodes = detectInvalidDiagnosticCodes;
exports.checkForInvalidDiagnosticCodes = checkForInvalidDiagnosticCodes;
const vscode_languageserver_1 = require("vscode-languageserver");
const error_codes_1 = require("./error-codes");
const node_types_1 = require("../utils/node-types");
const logger_1 = require("../logger");
// More precise regex to capture exact positions of code numbers
const DIAGNOSTIC_COMMENT_REGEX = /^#\s*@fish-lsp-(disable|enable)(?:-(next-line))?\s/;
function isPossibleDiagnosticComment(node) {
    if (!(0, node_types_1.isComment)(node))
        return false;
    return DIAGNOSTIC_COMMENT_REGEX.test(node.text.trim());
}
// Function to find codes with their positions
function findCodes(text) {
    // Find where the codes section starts (after the directive)
    const directiveMatch = text.match(/@fish-lsp-(?:disable|enable)(?:-next-line)?/); // remove leading comment
    if (!directiveMatch)
        return [];
    const codesStart = directiveMatch.index + directiveMatch[0].length;
    const codesSection = text.slice(codesStart);
    // Find all code tokens in the codes section
    const result = [];
    const codeRegex = /(\d+)/g;
    let match;
    while ((match = codeRegex.exec(codesSection)) !== null) {
        result.push({
            code: match[0],
            startIndex: codesStart + match.index,
        });
    }
    logger_1.logger.log('Found codes:', result, 'on text:', text);
    logger_1.logger.log('Directive:', directiveMatch);
    return result;
}
function detectInvalidDiagnosticCodes(node) {
    // Early return if not a diagnostic comment
    if (!(0, node_types_1.isComment)(node))
        return [];
    const text = node.text.trim();
    if (!DIAGNOSTIC_COMMENT_REGEX.test(text))
        return [];
    // Find all code numbers with their positions
    const codePositions = findCodes(text);
    const diagnostics = [];
    for (const { code, startIndex } of codePositions) {
        const codeNum = parseInt(code, 10);
        // Check if it's a valid error code
        if (isNaN(codeNum) || !error_codes_1.ErrorCodes.allErrorCodes.includes(codeNum)) {
            // Create diagnostic for this invalid code
            diagnostics.push({
                severity: vscode_languageserver_1.DiagnosticSeverity.Warning,
                range: {
                    start: {
                        line: node.startPosition.row,
                        character: node.startPosition.column + startIndex,
                    },
                    end: {
                        line: node.startPosition.row,
                        character: node.startPosition.column + startIndex + code.length,
                    },
                },
                message: `Invalid diagnostic code: '${code}'. Valid codes are: ${error_codes_1.ErrorCodes.allErrorCodes.join(', ')}.`,
                source: 'fish-lsp',
                code: 'invalidDiagnosticCode',
                data: {
                    node,
                    invalidCode: code,
                },
            });
        }
    }
    return diagnostics;
}
// Function to add to the validate.ts getDiagnostics function
function checkForInvalidDiagnosticCodes(node) {
    if (isPossibleDiagnosticComment(node)) {
        return detectInvalidDiagnosticCodes(node);
    }
    return [];
}
