"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ErrorNodeTypes = void 0;
exports.findErrorCause = findErrorCause;
exports.isExtraEnd = isExtraEnd;
exports.isZeroIndex = isZeroIndex;
exports.isSingleQuoteVariableExpansion = isSingleQuoteVariableExpansion;
exports.isAlias = isAlias;
exports.isUniversalDefinition = isUniversalDefinition;
exports.isSourceFilename = isSourceFilename;
exports.isDotSourceCommand = isDotSourceCommand;
exports.isTestCommandVariableExpansionWithoutString = isTestCommandVariableExpansionWithoutString;
exports.isConditionalStatement = isConditionalStatement;
exports.isFirstNodeInConditionalExecution = isFirstNodeInConditionalExecution;
exports.isConditionalWithoutQuietCommand = isConditionalWithoutQuietCommand;
exports.isVariableDefinitionWithExpansionCharacter = isVariableDefinitionWithExpansionCharacter;
exports.isMatchingCompleteOptionIsCommand = isMatchingCompleteOptionIsCommand;
exports.isArgparseWithoutEndStdin = isArgparseWithoutEndStdin;
exports.isFishLspDeprecatedVariableName = isFishLspDeprecatedVariableName;
exports.getDeprecatedFishLspMessage = getDeprecatedFishLspMessage;
const node_types_1 = require("../utils/node-types");
const tree_sitter_1 = require("../utils/tree-sitter");
const options_1 = require("../parsing/options");
const source_1 = require("../parsing/source");
exports.ErrorNodeTypes = {
    ['function']: 'end',
    ['while']: 'end',
    ['begin']: 'end',
    ['for']: 'end',
    ['if']: 'end',
    ['"']: '"',
    ["'"]: "'",
    ['{']: '}',
    ['[']: ']',
    ['(']: ')',
};
function isStartTokenType(str) {
    return ['function', 'while', 'if', 'for', 'begin', '[', '{', '(', "'", '"'].includes(str);
}
function findErrorCause(children) {
    const stack = [];
    for (const node of children) {
        if (isStartTokenType(node.type)) {
            const expectedEndToken = exports.ErrorNodeTypes[node.type];
            const matchIndex = stack.findIndex(item => item.type === expectedEndToken);
            if (matchIndex !== -1) {
                stack.splice(matchIndex, 1); // Remove the matched end token
            }
            else {
                stack.push({ node, type: expectedEndToken }); // Push the current node and expected end token to the stack
            }
        }
        else if (Object.values(exports.ErrorNodeTypes).includes(node.type)) {
            stack.push({ node, type: node.type }); // Track all end tokens
        }
    }
    // Return the first unmatched start token from the stack, if any
    return stack.length > 0 ? stack[0]?.node || null : null;
}
function isExtraEnd(node) {
    return node.type === 'command' && node.text === 'end';
}
function isZeroIndex(node) {
    return node.type === 'index' && node.text === '0';
}
function isSingleQuoteVariableExpansion(node) {
    if (node.type !== 'single_quote_string') {
        return false;
    }
    if (node.parent && (0, node_types_1.isCommandWithName)(node.parent, 'string')) {
        return false;
    }
    const variableRegex = /(?<!\\)\$\w+/; // Matches $variable, not preceded by a backslash
    return variableRegex.test(node.text);
}
function isAlias(node) {
    return (0, node_types_1.isCommandWithName)(node, 'alias');
}
function isUniversalDefinition(node) {
    const parent = node.parent;
    if (!parent)
        return false;
    if ((0, node_types_1.isCommandWithName)(parent, 'read') || (0, node_types_1.isCommandWithName)(parent, 'set')) {
        return (0, node_types_1.isMatchingOption)(node, options_1.Option.create('-U', '--universal'));
    }
    return false;
}
function isSourceFilename(node) {
    if ((0, source_1.isSourceCommandArgumentName)(node)) {
        const isExisting = (0, source_1.isExistingSourceFilenameNode)(node);
        if (!isExisting) {
            // check if the node is a variable expansion
            // if it is, do not through a diagnostic because we can't evaluate if this is a valid path
            // An example of this case:
            // for file in $__fish_data_dir/functions
            //     source $file # <--- we have no clue if this file exists
            // end
            if (node.type === 'variable_expansion') {
                return false;
            }
            // also skip something like `source '$file'`
            if ((0, node_types_1.isString)(node)) {
                return false;
            }
            return true;
        }
        return !isExisting;
    }
    return false;
}
function isDotSourceCommand(node) {
    if (node.parent && (0, node_types_1.isCommandWithName)(node.parent, '.')) {
        return node.parent.firstNamedChild?.equals(node) || false;
    }
    return false;
}
function isTestCommandVariableExpansionWithoutString(node) {
    const parent = node.parent;
    const previousSibling = node.previousSibling;
    if (!parent || !previousSibling)
        return false;
    if (!(0, node_types_1.isCommandWithName)(parent, 'test', '['))
        return false;
    if ((0, node_types_1.isMatchingOption)(previousSibling, options_1.Option.short('-n'), options_1.Option.short('-z'))) {
        return !(0, node_types_1.isString)(node) && !!parent.child(2) && parent.child(2).equals(node);
    }
    return false;
}
function isInsideStatementCondition(statement, node) {
    const conditionNode = statement.childForFieldName('condition');
    if (!conditionNode)
        return false;
    return (0, tree_sitter_1.isNodeWithinOtherNode)(node, conditionNode);
}
/**
 * util for collecting if conditional_statement commands
 * Necessary because there is two types of conditional statements:
 *    1.) if cmd_1 || cmd_2; ...; end;
 *    2.) if cmd_1; or cmd_2; ...; end;
 * Case two is handled by the if statement, checking for the parent type of conditional_execution
 * @param node - the current node to check (should be a command)
 * @returns true if the node is a conditional statement, otherwise false
 */
function isConditionalStatement(node) {
    if (!node.isNamed)
        return false;
    if (['\n', ';'].includes(node?.previousSibling?.type || ''))
        return false;
    let curr = node.parent;
    while (curr) {
        if (curr.type === 'conditional_execution') {
            curr = curr?.parent;
        }
        else if ((0, node_types_1.isIfOrElseIfConditional)(curr)) {
            return isInsideStatementCondition(curr, node);
        }
        else {
            break;
        }
    }
    return false;
}
/**
 * Check if a conditional_execution node starts with a conditional operator
 */
function checkConditionalStartsWith(node) {
    if (node.type === 'conditional_execution') {
        return node.text.startsWith('&&') || node.text.startsWith('||')
            || node.text.startsWith('and') || node.text.startsWith('or');
    }
    return false;
}
/**
 * Check if a command node is the first node in a conditional_execution
 */
function isFirstNodeInConditionalExecution(node) {
    if (!node.isNamed)
        return false;
    if (['\n', ';'].includes(node?.type || ''))
        return false;
    if (isConditionalStatement(node))
        return false;
    if (node.parent &&
        node.parent.type === 'conditional_execution' &&
        !checkConditionalStartsWith(node.parent)) {
        return node.parent.firstNamedChild?.equals(node) || false;
    }
    const next = node.nextNamedSibling;
    if (!next)
        return false;
    return next.type === 'conditional_execution' && checkConditionalStartsWith(next);
}
/**
 * Checks if a command has a command substitution. For example,
 *
 *   ```fish
 *   if set -l fishdir (status fish-path | string match -vr /bin/)
 *       echo $fishdir
 *   end
 *   ```
 *
 * @param node - the current node to check (should be a `set` command)
 * @returns true if the command has a command substitution, otherwise false
 */
function hasCommandSubstitution(node) {
    return node.childrenForFieldName('argument').filter(c => c.type === 'command_substitution').length > 0;
}
/**
 * Check if -q,--quiet/--query flags are present for commands which follow an `if/else if` conditional statement
 * @param node - the current node to check (should be a command)
 * @returns true if the command is a conditional statement without -q,--quiet/--query flags, otherwise false
 */
function isConditionalWithoutQuietCommand(node) {
    // if (!isCommand(node)) return false;
    if (!(0, node_types_1.isCommandWithName)(node, 'command', 'type', 'read', 'set', 'string', 'abbr', 'builtin', 'functions', 'jobs'))
        return false;
    if (!isConditionalStatement(node) && !isFirstNodeInConditionalExecution(node))
        return false;
    // skip `set` commands with command substitution
    if ((0, node_types_1.isCommandWithName)(node, 'set') && hasCommandSubstitution(node)) {
        return false;
    }
    const flags = node?.childrenForFieldName('argument')
        .filter(n => (0, node_types_1.isMatchingOption)(n, options_1.Option.create('-q', '--quiet'))
        || (0, node_types_1.isMatchingOption)(n, options_1.Option.create('-q', '--query'))) || [];
    return flags.length === 0;
}
function isVariableDefinitionWithExpansionCharacter(node) {
    if (node.parent && (0, node_types_1.isCommandWithName)(node.parent, 'set', 'read')) {
        const definition = (0, tree_sitter_1.getChildNodes)(node.parent).filter(n => !(0, node_types_1.isCommand)(n) && !(0, node_types_1.isCommandName)(n) && !(0, node_types_1.isOption)(n)).shift();
        return (node.type === 'variable_expansion' || node.text.startsWith('$')) && definition?.equals(node);
    }
    return false;
}
function isMatchingCompleteOptionIsCommand(node) {
    return (0, node_types_1.isMatchingOption)(node, options_1.Option.create('-n', '--condition').withValue())
        || (0, node_types_1.isMatchingOption)(node, options_1.Option.create('-a', '--arguments').withValue())
        || (0, node_types_1.isMatchingOption)(node, options_1.Option.create('-c', '--command').withValue());
}
function isArgparseWithoutEndStdin(node) {
    if (!(0, node_types_1.isCommandWithName)(node, 'argparse'))
        return false;
    const endStdin = (0, tree_sitter_1.getChildNodes)(node).find(n => (0, node_types_1.isEndStdinCharacter)(n));
    if (!endStdin)
        return true;
    return false;
}
function isFishLspDeprecatedVariableName(node) {
    if ((0, node_types_1.isVariableDefinitionName)(node)) {
        return node.text === 'fish_lsp_logfile';
    }
    if (node.type === 'variable_name') {
        return node.text === 'fish_lsp_logfile';
    }
    return node.text === 'fish_lsp_logfile';
}
function getDeprecatedFishLspMessage(node) {
    switch (node.text) {
        case 'fish_lsp_logfile':
            return `REPLACE \`${node.text}\` with \`fish_lsp_log_file\``;
        default:
            return '';
    }
}
