"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CompletionSymbol = exports.CompleteOptions = void 0;
exports.isCompletionCommandDefinition = isCompletionCommandDefinition;
exports.isCompletionDefinitionWithName = isCompletionDefinitionWithName;
exports.isCompletionSymbolShort = isCompletionSymbolShort;
exports.isCompletionSymbolLong = isCompletionSymbolLong;
exports.isCompletionSymbolOld = isCompletionSymbolOld;
exports.isCompletionSymbol = isCompletionSymbol;
exports.isCompletionSymbolVerbose = isCompletionSymbolVerbose;
exports.getCompletionSymbol = getCompletionSymbol;
exports.groupCompletionSymbolsTogether = groupCompletionSymbolsTogether;
exports.processCompletion = processCompletion;
const node_types_1 = require("../utils/node-types");
const options_1 = require("./options");
const tree_sitter_1 = require("../utils/tree-sitter");
const vscode_languageserver_1 = require("vscode-languageserver");
exports.CompleteOptions = [
    options_1.Option.create('-c', '--command').withValue(),
    options_1.Option.create('-p', '--path'),
    options_1.Option.create('-e', '--erase'),
    options_1.Option.create('-s', '--short-option').withValue(),
    options_1.Option.create('-l', '--long-option').withValue(),
    options_1.Option.create('-o', '--old-option').withValue(),
    options_1.Option.create('-a', '--arguments').withValue(),
    options_1.Option.create('-k', '--keep-order'),
    options_1.Option.create('-f', '--no-files'),
    options_1.Option.create('-F', '--force-files'),
    options_1.Option.create('-r', '--require-parameter'),
    options_1.Option.create('-x', '--exclusive'),
    options_1.Option.create('-d', '--description').withValue(),
    options_1.Option.create('-w', '--wraps').withValue(),
    options_1.Option.create('-n', '--condition').withValue(),
    options_1.Option.create('-C', '--do-complete').withValue(),
    options_1.Option.long('--escape').withValue(),
    options_1.Option.create('-h', '--help'),
];
function isCompletionCommandDefinition(node) {
    return (0, node_types_1.isCommandWithName)(node, 'complete');
}
function isCompletionDefinitionWithName(node, name, doc) {
    if (node.parent && isCompletionCommandDefinition(node.parent)) {
        const symbol = getCompletionSymbol(node.parent, doc);
        return symbol?.commandName === name && isCompletionSymbol(node);
    }
    return false;
}
function isCompletionSymbolShort(node) {
    if (node.parent && isCompletionCommandDefinition(node.parent)) {
        return node.previousSibling && (0, options_1.isMatchingOption)(node.previousSibling, options_1.Option.create('-s', '--short-option'));
    }
    return false;
}
function isCompletionSymbolLong(node) {
    if (node.parent && isCompletionCommandDefinition(node.parent)) {
        return node.previousSibling && (0, options_1.isMatchingOption)(node.previousSibling, options_1.Option.create('-l', '--long-option'));
    }
    return false;
}
function isCompletionSymbolOld(node) {
    if (node.parent && isCompletionCommandDefinition(node.parent)) {
        return node.previousSibling && (0, options_1.isMatchingOption)(node.previousSibling, options_1.Option.create('-o', '--old-option'));
    }
    return false;
}
function isCompletionSymbol(node) {
    return isCompletionSymbolShort(node)
        || isCompletionSymbolLong(node)
        || isCompletionSymbolOld(node);
}
class CompletionSymbol {
    optionType;
    commandName;
    node;
    description;
    condition;
    requireParameter;
    argumentNames;
    exclusive;
    doc;
    constructor(optionType = '', commandName = '', node = null, description = '', condition = '', requireParameter = false, argumentNames = '', exclusive = false, doc) {
        this.optionType = optionType;
        this.commandName = commandName;
        this.node = node;
        this.description = description;
        this.condition = condition;
        this.requireParameter = requireParameter;
        this.argumentNames = argumentNames;
        this.exclusive = exclusive;
        this.doc = doc;
    }
    /**
     * Initialize the VerboseCompletionSymbol with empty values.
     */
    static createEmpty() {
        return new CompletionSymbol();
    }
    /**
     * util for building a VerboseCompletionSymbol
     */
    static create({ optionType = '', commandName = '', node = null, description = '', condition = '', requireParameter = false, argumentNames = '', exclusive = false, }) {
        return new this(optionType, commandName, node, description, condition, requireParameter, argumentNames, exclusive);
    }
    /**
     * If the node is not found, we don't have a valid VerboseCompletionSymbol.
     */
    isEmpty() {
        return this.node === null;
    }
    /**
     * Type Guard that our node & its parent are defined,
     * therefore we have found a valid VerboseCompletionSymbol.
     */
    isNonEmpty() {
        return this.node !== null && this.parent !== null;
    }
    /**
     * Getter (w/ type guarding) to retrieve the CompletionSymbol.node.parent
     * Removes the pattern of null checking a CompletionSymbol.node.parent
     */
    get parent() {
        if (this.node) {
            return this.node.parent;
        }
        return null;
    }
    /**
     * Getter (w/ type guarding) to retrieve the CompletionSymbol.node.text
     * Removes the pattern of null checking a CompletionSymbol.node
     */
    get text() {
        if (this.isNonEmpty()) {
            return this.node.text;
        }
        return '';
    }
    /**
     * Check if the option is a short option: `-s <flag>` or `--short-option <flag>`.
     */
    isShort() {
        return this.optionType === 'short';
    }
    /**
     * Check if the option is a long option: `-l <flag>` or `--long-option <flag>`.
     */
    isLong() {
        return this.optionType === 'long';
    }
    /**
     * Check if the option is an old option: `-o <flag>` or `--old-option <flag>`.
     */
    isOld() {
        return this.optionType === 'old';
    }
    /**
     * Check if one option is a pair of another option.
     * ```fish
     * complete -c foo -s h -l help # 'h' <--> 'help' are pairs
     * ```
     */
    isCorrespondingOption(other) {
        if (!this.isNonEmpty() || !other.isNonEmpty()) {
            return false;
        }
        return this.parent.equals(other.parent)
            && this.commandName === other.commandName
            && this.optionType !== other.optionType;
    }
    /**
     * Return the `-f`/`--flag`/`-flag` string
     */
    toFlag() {
        if (!this.isNonEmpty()) {
            return '';
        }
        switch (this.optionType) {
            case 'short':
            case 'old':
                return `-${this.node.text}`;
            case 'long':
                return `--${this.node.text}`;
            default:
                return '';
        }
    }
    /**
     * return the commandName and the flag as a string
     */
    toUsage() {
        if (!this.isNonEmpty()) {
            return '';
        }
        return `${this.commandName} ${this.toFlag()}`;
    }
    /**
     * return the usage, with the description in a trailing comment
     */
    toUsageVerbose() {
        if (!this.isNonEmpty()) {
            return '';
        }
        return `${this.commandName} ${this.toFlag()} # ${this.description}`;
    }
    /**
     * check if the symbol inside a globally defined `argparse o/opt -- $argv` matches
     * this VerboseCompletionSymbol
     */
    equalsArgparse(symbol) {
        if (symbol.fishKind !== 'ARGPARSE' || !symbol.parent) {
            return false;
        }
        const commandName = symbol.parent.name;
        const symbolName = symbol.argparseFlagName;
        return this.commandName === commandName
            && this.node?.text === symbolName;
    }
    /**
     * Check if our CompletionSymbol.node === the node passed in
     */
    equalsNode(n) {
        return this.node?.equals(n);
    }
    /**
     * check if our CompletionSymbol.commandName === the commandName passed in
     */
    hasCommandName(name) {
        return this.commandName === name;
    }
    /**
     * A test utility for easily getting a completion flag
     */
    isMatchingRawOption(...opts) {
        const flag = this.toFlag();
        for (const opt of opts) {
            if (flag === opt) {
                return true;
            }
        }
        return false;
    }
    /**
     * utility to get the range of the node
     */
    getRange() {
        if (this.isNonEmpty()) {
            return (0, tree_sitter_1.getRange)(this.node);
        }
        return null;
    }
    /**
     * Create a Location from the current CompletionSymbol
     */
    toLocation() {
        return vscode_languageserver_1.Location.create(this.doc?.uri || '', this.getRange());
    }
}
exports.CompletionSymbol = CompletionSymbol;
function isCompletionSymbolVerbose(node, doc) {
    if (isCompletionSymbol(node) || !node.parent) {
        return true;
    }
    if (node.parent && isCompletionCommandDefinition(node.parent)) {
        const symbol = getCompletionSymbol(node, doc);
        return symbol?.isNonEmpty() || false;
    }
    return false;
}
/**
 * Create a VerboseCompletionSymbol from a SyntaxNode, for any SyntaxNode passed in.
 * Calling this function will need to check if `result.isEmpty()` or `result.isNonEmpty()`
 * @param node any syntax node, preferably one that is a child of a `complete` node (not required though)
 * @returns {CompletionSymbol} `result.isEmpty()` when not found, `result.isNonEmpty()` when `isCompletionSymbolVerbose(node)` is found
 */
function getCompletionSymbol(node, doc) {
    const result = CompletionSymbol.createEmpty();
    if (!isCompletionSymbol(node) || !node.parent) {
        return result;
    }
    switch (true) {
        case isCompletionSymbolShort(node):
            result.optionType = 'short';
            break;
        case isCompletionSymbolLong(node):
            result.optionType = 'long';
            break;
        case isCompletionSymbolOld(node):
            result.optionType = 'old';
            break;
        default:
            break;
    }
    result.node = node;
    const parent = node.parent;
    const children = parent.childrenForFieldName('argument');
    result.doc = doc;
    children.forEach((child, idx) => {
        if (idx === 0)
            return;
        if ((0, options_1.isMatchingOption)(child, options_1.Option.create('-r', '--require-parameter'))) {
            result.requireParameter = true;
        }
        if ((0, options_1.isMatchingOption)(child, options_1.Option.create('-x', '--exclusive'))) {
            result.exclusive = true;
        }
        const prev = child.previousSibling;
        if (!prev)
            return;
        if ((0, options_1.isMatchingOption)(prev, options_1.Option.create('-c', '--command'))) {
            result.commandName = child.text;
        }
        if ((0, options_1.isMatchingOption)(prev, options_1.Option.create('-d', '--description'))) {
            result.description = (0, node_types_1.isString)(child) ? child.text.slice(1, -1) : child.text;
        }
        if ((0, options_1.isMatchingOption)(prev, options_1.Option.create('-n', '--condition'))) {
            result.condition = child.text;
        }
        if ((0, options_1.isMatchingOption)(prev, options_1.Option.create('-a', '--arguments'))) {
            result.argumentNames = child.text;
        }
    });
    return result;
}
function groupCompletionSymbolsTogether(...symbols) {
    const storedSymbols = new Set();
    const groupedSymbols = [];
    symbols.forEach((symbol) => {
        if (storedSymbols.has(symbol.text)) {
            return;
        }
        const newGroup = [symbol];
        const matches = symbols.filter((s) => s.isCorrespondingOption(symbol));
        matches.forEach((s) => {
            storedSymbols.add(s.text);
            newGroup.push(s);
        });
        groupedSymbols.push(newGroup);
    });
    return groupedSymbols;
}
function processCompletion(document, node) {
    const result = [];
    for (const child of (0, tree_sitter_1.getChildNodes)(node)) {
        if (isCompletionCommandDefinition(node)) {
            const newSymbol = getCompletionSymbol(child, document);
            if (newSymbol)
                result.push(newSymbol);
        }
    }
    return result;
}
