"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.FunctionOptions = void 0;
exports.findFunctionDefinitionChildren = findFunctionDefinitionChildren;
exports.processArgvDefinition = processArgvDefinition;
exports.isFunctionDefinitionName = isFunctionDefinitionName;
exports.isFunctionVariableDefinitionName = isFunctionVariableDefinitionName;
exports.processFunctionDefinition = processFunctionDefinition;
const options_1 = require("./options");
const symbol_1 = require("./symbol");
const node_types_1 = require("../utils/node-types");
const snippets_1 = require("../utils/snippets");
const definition_scope_1 = require("../utils/definition-scope");
const translation_1 = require("../utils/translation");
const tree_sitter_1 = require("../utils/tree-sitter");
exports.FunctionOptions = [
    options_1.Option.create('-a', '--argument-names').withMultipleValues(),
    options_1.Option.create('-d', '--description').withValue(),
    options_1.Option.create('-w', '--wraps').withValue(),
    options_1.Option.create('-e', '--on-event').withValue(),
    options_1.Option.create('-v', '--on-variable').withValue(),
    options_1.Option.create('-j', '--on-job-exit').withValue(),
    options_1.Option.create('-p', '--on-process-exit').withValue(),
    options_1.Option.create('-s', '--on-signal').withValue(),
    options_1.Option.create('-S', '--no-scope-shadowing'),
    options_1.Option.create('-V', '--inherit-variable').withValue(),
];
function isFunctionDefinition(node) {
    return node.type === 'function_definition';
}
/**
 * Util to find all the arguments of a function_definition node
 *
 * function foo -a bar baz -V foobar -d '...' -w '...' --on-event '...'
 *               ^  ^   ^   ^  ^      ^  ^    ^   ^     ^         ^
 *               all of these nodes would be returned in the SyntaxNode[] array
 * @param node the function_definition node
 * @returns SyntaxNode[] of all the arguments to the function_definition
 */
function findFunctionDefinitionChildren(node) {
    return node.childrenForFieldName('option').filter(n => !(0, node_types_1.isEscapeSequence)(n) && !(0, node_types_1.isNewline)(n));
}
/**
 * Get argv definition for fish shell script files (non auto-loaded files)
 */
function processArgvDefinition(document, node) {
    if (!document.isAutoloaded() && node.type === 'program') {
        return [
            symbol_1.FishSymbol.fromObject({
                name: 'argv',
                node: node,
                focusedNode: node.firstChild,
                fishKind: symbol_1.FishSymbolKindMap.variable,
                uri: document.uri,
                detail: snippets_1.PrebuiltDocumentationMap.getByName('argv').pop()?.description || 'the list of arguments passed to the function',
                scope: definition_scope_1.DefinitionScope.create(node, 'local'),
                selectionRange: {
                    start: { line: 0, character: 0 },
                    end: { line: 0, character: 0 },
                },
                range: (0, tree_sitter_1.getRange)(node),
                children: [],
            }),
        ];
    }
    return [];
}
/**
 * checks if a node is the function name of a function definition
 * function foo
 *          ^--- here
 */
function isFunctionDefinitionName(node) {
    if (!node.parent || !isFunctionDefinition(node.parent))
        return false;
    return !!node.parent.firstNamedChild && node.parent.firstNamedChild.equals(node);
}
/**
 * checks if a node is the variable name of a function definition
 * function foo --argument-names bar baz --inherit-variable foobar
 *                               ^   ^                       ^
 *                               |   |                       |
 *                               Could be any of these nodes above
 * Currently doesn't check for `--on-variable`, because it should be inherited
 */
function isFunctionVariableDefinitionName(node) {
    if (!node.parent || !isFunctionDefinition(node.parent))
        return false;
    const { variableNodes } = findFunctionVariableArguments(node.parent);
    const definitionNode = variableNodes.find(n => n.equals(node));
    return !!definitionNode && definitionNode.equals(node);
}
/**
 * Find all the function_definition variables that are defined in the function header
 *
 * The `flagsSet` property contains all the nodes that were found to be variable names,
 * with the flag that was used to define them.
 *
 * @param node the function_definition node
 * @returns Object containing the defined SyntaxNode[] and OptionValueMatch[] flags set
 */
function findFunctionVariableArguments(node) {
    const variableNodes = [];
    const focused = node.childrenForFieldName('option').filter(n => !(0, node_types_1.isEscapeSequence)(n) && !(0, node_types_1.isNewline)(n));
    const flagsSet = (0, options_1.findOptionsSet)(focused, exports.FunctionOptions);
    for (const flag of flagsSet) {
        const { option, value: focused } = flag;
        switch (true) {
            case option.isOption('-a', '--argument-names'):
            case option.isOption('-V', '--inherit-variable'):
                // case option.isOption('-v', '--on-variable'):
                variableNodes.push(focused);
                break;
            default:
                break;
        }
    }
    return {
        variableNodes,
        flagsSet,
    };
}
/**
 * Process a function definition node and return the corresponding FishSymbol[]
 * for the function and its arguments. Includes argv as a child, along with any
 * flags that create function scoped variables + any children nodes are stored as well.
 */
function processFunctionDefinition(document, node, children = []) {
    if (!isFunctionDefinition(node))
        return [];
    const autoloadScope = (0, translation_1.isAutoloadedUriLoadsFunctionName)(document);
    const focusedNode = node.firstNamedChild;
    const isGlobal = autoloadScope(focusedNode) ? 'global' : 'local';
    if (!focusedNode)
        return [];
    const functionSymbol = symbol_1.FishSymbol.create(focusedNode.text, node, focusedNode, symbol_1.FishSymbolKindMap.function, document.uri, node.text, definition_scope_1.DefinitionScope.create(node.parent, isGlobal));
    const focused = node.childrenForFieldName('option').filter(n => !(0, node_types_1.isEscapeSequence)(n) && !(0, node_types_1.isNewline)(n));
    functionSymbol.addChildren(symbol_1.FishSymbol.create('argv', node, node.firstNamedChild, symbol_1.FishSymbolKindMap.function_variable, document.uri, snippets_1.PrebuiltDocumentationMap.getByName('argv').pop()?.description || 'the list of arguments passed to the function', definition_scope_1.DefinitionScope.create(node, 'local')));
    if (!focused)
        return [functionSymbol];
    const { flagsSet } = findFunctionVariableArguments(node);
    for (const flag of flagsSet) {
        const { option, value: focused } = flag;
        switch (true) {
            case option.isOption('-a', '--argument-names'):
            case option.isOption('-V', '--inherit-variable'):
                // case option.isOption('-v', '--on-variable'):
                functionSymbol.addChildren(symbol_1.FishSymbol.create(focused.text, node, focused, symbol_1.FishSymbolKindMap.function_variable, document.uri, focused.text, definition_scope_1.DefinitionScope.create(node, 'local')));
                break;
            default:
                break;
        }
    }
    return [functionSymbol.addChildren(...children)];
}
