"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.unindentNestedSyntaxNode = unindentNestedSyntaxNode;
exports.isAutoloadedPath = isAutoloadedPath;
exports.getAutoloadType = getAutoloadType;
exports.createDetail = createDetail;
const vscode_languageserver_1 = require("vscode-languageserver");
const symbol_1 = require("./symbol");
const markdown_builder_1 = require("../utils/markdown-builder");
const options_1 = require("./options");
const function_1 = require("./function");
const node_types_1 = require("../utils/node-types");
const translation_1 = require("../utils/translation");
const snippets_1 = require("../utils/snippets");
const set_1 = require("./set");
const alias_1 = require("./alias");
const env_manager_1 = require("../utils/env-manager");
const logger_1 = require("../logger");
// IF YOU ARE READING THIS FILE, PLEASE FEEL FREE TO REFACTOR IT (sorry my brain is fried)
/**
 * Since a SyntaxNode's text could equal something like:
 * ```fish
 * # assume we are indented one level, (if_statement wont have leading spaces)
 * if true
 *         echo "Hello, world!"
 *     end
 * ```
 * We want to remove a single indentation level from the text, after the first line.
 * @param node The SyntaxNode to unindent
 * @returns The unindented text of the SyntaxNode (the last line's indentation amount will be how much is removed from the rest of the lines)
 */
function unindentNestedSyntaxNode(node) {
    const lines = node.text.split('\n');
    if (lines.length > 1) {
        const lastLine = node.lastChild?.startPosition.column || 0;
        return lines
            .map(line => line.replace(' '.repeat(lastLine), ''))
            .join('\n')
            .trimEnd();
    }
    return node.text;
}
function getSymbolKind(symbol) {
    const kind = symbol_1.fishSymbolKindToSymbolKind[symbol.fishKind];
    switch (kind) {
        case vscode_languageserver_1.SymbolKind.Variable:
            return 'variable';
        case vscode_languageserver_1.SymbolKind.Function:
            return 'function';
        default:
            return '';
    }
}
/**
 * Checks if a file path is within any autoloaded fish directories
 *
 * @param uriOrPath The URI or filesystem path to check
 * @param type Optional specific autoload type to check for (e.g., 'functions', 'completions')
 * @returns True if the path is within an autoloaded directory, false otherwise
 */
function isAutoloadedPath(uriOrPath, type) {
    // Convert URI to path if necessary
    const path = uriOrPath.startsWith('file://') ? (0, translation_1.uriToPath)(uriOrPath) : uriOrPath;
    // Get all autoloaded variables from the environment
    const autoloadedKeys = env_manager_1.env.getAutoloadedKeys();
    for (const key of autoloadedKeys) {
        // Skip if we're looking for a specific type and this key doesn't match
        if (type && !key.toLowerCase().includes(type.toLowerCase())) {
            continue;
        }
        const values = env_manager_1.env.getAsArray(key);
        for (const value of values) {
            if (path.startsWith(value)) {
                return true;
            }
        }
    }
    return false;
}
/**
 * Gets the autoload type of a path if it is autoloaded
 *
 * @param uriOrPath The URI or filesystem path to check
 * @returns The identified autoload type ('functions', 'completions', 'conf.d', etc.) or empty string if not autoloaded
 */
function getAutoloadType(uriOrPath) {
    // Convert URI to path if necessary
    const path = uriOrPath.startsWith('file://') ? (0, translation_1.uriToPath)(uriOrPath) : uriOrPath;
    // Common autoload types to check for
    const autoloadTypes = ['functions', 'completions', 'conf.d', 'config'];
    // Check path for these common types
    for (const type of autoloadTypes) {
        if (path.includes(`/fish/${type}`)) {
            return type;
        }
        // Special case for config.fish
        if (type === 'config' && path.endsWith('config.fish')) {
            return 'config';
        }
    }
    // If no specific type was found but the path is autoloaded, return a generic indicator
    if (isAutoloadedPath(path)) {
        return 'autoloaded';
    }
    return '';
}
function buildFunctionDetail(symbol) {
    const { name, node, fishKind } = symbol;
    if (fishKind === 'ALIAS') {
        return alias_1.FishAlias.buildDetail(node);
    }
    const options = (0, options_1.findOptions)((0, function_1.findFunctionDefinitionChildren)(node), function_1.FunctionOptions);
    const descriptionOption = options.found.find(option => option.option.isOption('-d', '---description'));
    const description = [`(${markdown_builder_1.md.bold('function')}) ${markdown_builder_1.md.inlineCode(name)}`];
    if (descriptionOption && descriptionOption.value) {
        description.push((0, node_types_1.isString)(descriptionOption.value)
            ? descriptionOption.value.text.slice(1, -1)
            : descriptionOption.value.text);
    }
    description.push(markdown_builder_1.md.separator());
    const scope = [];
    if (isAutoloadedPath(symbol.uri)) {
        scope.push('autoloaded');
    }
    if (symbol.isGlobal()) {
        scope.push('globally scoped');
    }
    if (scope.length > 0) {
        description.push(scope.join(', '));
    }
    description.push(`located in file: ${markdown_builder_1.md.inlineCode((0, translation_1.uriToReadablePath)(symbol.uri))}`);
    description.push(markdown_builder_1.md.separator());
    const prebuilt = snippets_1.PrebuiltDocumentationMap.getByType('command').find(c => c.name === name);
    if (prebuilt) {
        description.push(prebuilt.description);
        description.push(markdown_builder_1.md.separator());
    }
    description.push(markdown_builder_1.md.codeBlock('fish', unindentNestedSyntaxNode(node)));
    const argumentNamesOption = options.found.filter(option => option.option.isOption('-a', '--argument-names'));
    if (argumentNamesOption && argumentNamesOption.length) {
        const functionCall = [name];
        for (const arg of argumentNamesOption) {
            functionCall.push(arg.value.text);
        }
        description.push(markdown_builder_1.md.separator());
        description.push(markdown_builder_1.md.codeBlock('fish', functionCall.join(' ')));
    }
    return description.join(markdown_builder_1.md.newline());
}
function isVariableArgumentNamed(node, name) {
    if (node.type !== 'function_definition')
        return '';
    const children = (0, function_1.findFunctionDefinitionChildren)(node);
    if ((0, options_1.findOptions)(children, function_1.FunctionOptions).found
        .filter(flag => flag.option.isOption('-a', '--argument-names'))
        .some(flag => flag.value.text === name)) {
        return true;
    }
    return false;
}
function getArgumentNamesIndexString(node, name) {
    if (node?.type && node?.type !== 'function_definition')
        return '';
    const children = (0, function_1.findFunctionDefinitionChildren)(node);
    // const resultFlags: string[] = [];
    const index = (0, options_1.findOptions)(children, function_1.FunctionOptions).found
        .filter(flag => flag.option.isOption('-a', '--argument-names'))
        .findIndex((flag) => flag.value.text === name);
    const argvStr = '$argv[' + (index + 1) + ']';
    return `${markdown_builder_1.md.italic('named argument')}: ${markdown_builder_1.md.inlineCode(argvStr)}`;
}
function buildVariableDetail(symbol) {
    const { name, node, uri, fishKind } = symbol;
    if (!node)
        return '';
    const description = [`(${markdown_builder_1.md.bold('variable')}) ${markdown_builder_1.md.inlineCode(name)}`];
    // add short info about variable
    description.push(markdown_builder_1.md.separator());
    if (fishKind === 'SET' || fishKind === 'READ') {
        const setModifiers = set_1.SetModifiers.filter(option => option.equalsRawLongOption('--universal', '--global', '--function', '--local', '--export', '--unexport'));
        const options = (0, options_1.findOptions)(node.childrenForFieldName('argument'), setModifiers);
        const modifier = options.found.find(o => o.option.equalsRawOption('-U', '-g', '-f', '-l', '-x', '-u'));
        if (modifier) {
            description.push((0, set_1.setModifierDetailDescriptor)(node));
        }
    }
    else if (fishKind === 'ARGPARSE') {
        description.push('locally scoped');
    }
    else if (node && isVariableArgumentNamed(node, name)) {
        try {
            const result = getArgumentNamesIndexString(node, name);
            description.push(result);
        }
        catch (e) {
            logger_1.logger.error('ERROR: building variable detail', e);
        }
    }
    // add location
    description.push(`located in file: ${markdown_builder_1.md.inlineCode((0, translation_1.uriToReadablePath)(uri))}`);
    // add prebuilt documentation if available
    const prebuilt = snippets_1.PrebuiltDocumentationMap.getByType('variable').find(c => c.name === name);
    if (prebuilt) {
        description.push(markdown_builder_1.md.separator());
        description.push(prebuilt.description);
    }
    description.push(markdown_builder_1.md.separator());
    // add code block of entire region
    description.push(markdown_builder_1.md.codeBlock('fish', unindentNestedSyntaxNode(node)));
    // add trailing `cmd --arg`, `cmd $argv`, `func $argv` examples
    const scopeCommand = symbol.scope.scopeNode?.type === 'program'
        ? `${(0, translation_1.uriToReadablePath)(uri)}`
        : `${symbol.scope.scopeNode?.firstNamedChild?.text}` || `${symbol.node}`;
    if (fishKind === 'ARGPARSE') {
        const argumentNamesOption = symbol.name.slice('_flag_'.length).replace(/_/g, '-');
        if (argumentNamesOption.length > 1) {
            description.push(markdown_builder_1.md.separator());
            description.push(markdown_builder_1.md.codeBlock('fish', `${scopeCommand} --${argumentNamesOption}`));
        }
        else if (argumentNamesOption.length === 1) {
            description.push(markdown_builder_1.md.separator());
            description.push(markdown_builder_1.md.codeBlock('fish', `${scopeCommand} -${argumentNamesOption}`));
        }
    }
    else if (name === 'argv') {
        description.push(markdown_builder_1.md.separator());
        description.push(markdown_builder_1.md.codeBlock('fish', `${scopeCommand} $argv`));
    }
    else if (node.type === 'function_definition') {
        const children = (0, function_1.findFunctionDefinitionChildren)(node);
        const resultFlags = [];
        (0, options_1.findOptions)(children, function_1.FunctionOptions).found
            .filter(flag => flag.option.isOption('-a', '--argument-names'))
            .forEach((flag, idx) => {
            if (flag.value.text === name)
                resultFlags.push(flag.value.text);
            else
                resultFlags.push(`\$argv[${idx + 1}]`);
        });
        if (resultFlags.length) {
            description.push(markdown_builder_1.md.separator());
            description.push(markdown_builder_1.md.codeBlock('fish', `${scopeCommand} ${resultFlags.join(' ')}`));
        }
    }
    return description.join(markdown_builder_1.md.newline());
}
function createDetail(symbol) {
    if (symbol.fishKind === 'EXPORT')
        return symbol.detail.toString();
    const symbolKind = getSymbolKind(symbol);
    if (symbolKind === '')
        return symbol.detail;
    if (symbolKind === 'function') {
        return buildFunctionDetail(symbol);
    }
    if (symbolKind === 'variable') {
        return buildVariableDetail(symbol);
    }
    return symbol.detail.toString();
}
