"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.FishSymbol = exports.fromFishSymbolKindToSymbolKind = exports.SetModifierToScopeTag = exports.fishSymbolKindToSymbolKind = exports.FishSymbolKindMap = void 0;
exports.getLocalSymbols = getLocalSymbols;
exports.getGlobalSymbols = getGlobalSymbols;
exports.isSymbol = isSymbol;
exports.filterLastPerScopeSymbol = filterLastPerScopeSymbol;
exports.findLocalLocations = findLocalLocations;
exports.findMatchingLocations = findMatchingLocations;
exports.removeLocalSymbols = removeLocalSymbols;
exports.formatFishSymbolTree = formatFishSymbolTree;
exports.processNestedTree = processNestedTree;
const vscode_languageserver_1 = require("vscode-languageserver");
const tree_sitter_1 = require("../utils/tree-sitter");
const set_1 = require("./set");
const read_1 = require("./read");
const function_1 = require("./function");
const for_1 = require("./for");
const argparse_1 = require("./argparse");
const options_1 = require("./options");
const alias_1 = require("./alias");
const symbol_detail_1 = require("./symbol-detail");
const config_1 = require("../config");
const flatten_1 = require("../utils/flatten");
const translation_1 = require("../utils/translation");
const node_types_1 = require("../utils/node-types");
const file_operations_1 = require("../utils/file-operations");
const export_1 = require("./export");
exports.FishSymbolKindMap = {
    ['argparse']: 'ARGPARSE',
    ['function']: 'FUNCTION',
    ['alias']: 'ALIAS',
    ['complete']: 'COMPLETE',
    ['set']: 'SET',
    ['read']: 'READ',
    ['for']: 'FOR',
    ['variable']: 'VARIABLE',
    ['function_variable']: 'FUNCTION_VARIABLE',
    ['export']: 'EXPORT',
};
exports.fishSymbolKindToSymbolKind = {
    ['ARGPARSE']: vscode_languageserver_1.SymbolKind.Variable,
    ['FUNCTION']: vscode_languageserver_1.SymbolKind.Function,
    ['ALIAS']: vscode_languageserver_1.SymbolKind.Function,
    ['COMPLETE']: vscode_languageserver_1.SymbolKind.Interface,
    ['SET']: vscode_languageserver_1.SymbolKind.Variable,
    ['READ']: vscode_languageserver_1.SymbolKind.Variable,
    ['FOR']: vscode_languageserver_1.SymbolKind.Variable,
    ['VARIABLE']: vscode_languageserver_1.SymbolKind.Variable,
    ['FUNCTION_VARIABLE']: vscode_languageserver_1.SymbolKind.Variable,
    ['EXPORT']: vscode_languageserver_1.SymbolKind.Variable,
};
const SetModifierToScopeTag = (modifier) => {
    switch (true) {
        case modifier.isOption('-U', '--universal'):
            return 'universal';
        case modifier.isOption('-g', '--global'):
            return 'global';
        case modifier.isOption('-f', '--function'):
            return 'function';
        case modifier.isOption('-l', '--local'):
            return 'local';
        default:
            return 'local';
    }
};
exports.SetModifierToScopeTag = SetModifierToScopeTag;
const fromFishSymbolKindToSymbolKind = (kind) => exports.fishSymbolKindToSymbolKind[kind];
exports.fromFishSymbolKindToSymbolKind = fromFishSymbolKindToSymbolKind;
class FishSymbol {
    children = [];
    aliasedNames = [];
    constructor(obj) {
        this.name = obj.name || obj.focusedNode.text;
        this.kind = (0, exports.fromFishSymbolKindToSymbolKind)(obj.fishKind);
        this.fishKind = obj.fishKind;
        this.uri = obj.uri;
        this.range = obj.range || (0, tree_sitter_1.getRange)(obj.node);
        this.selectionRange = obj.selectionRange || (0, tree_sitter_1.getRange)(obj.focusedNode);
        this.node = obj.node;
        this.focusedNode = obj.focusedNode;
        this.scope = obj.scope;
        this.children = obj.children;
        this.children.forEach(child => {
            child.parent = this;
        });
        this.detail = obj.detail;
        this.setupDetail();
    }
    setupDetail() {
        this.detail = (0, symbol_detail_1.createDetail)(this);
    }
    static create(name, node, focusedNode, fishKind, uri, detail, scope, children = []) {
        return new this({
            name: name || focusedNode.text,
            fishKind,
            uri,
            detail,
            node,
            focusedNode,
            scope,
            children,
        });
    }
    static fromObject(obj) {
        return new this(obj);
    }
    addChildren(...children) {
        this.children.push(...children);
        children.forEach(child => {
            child.parent = this;
        });
        return this;
    }
    addAliasedNames(...names) {
        this.aliasedNames.push(...names);
        return this;
    }
    nameEqualsNodeText(node) {
        return this.name === node.text;
    }
    get argparseFlagName() {
        return this.name.replace(/^_flag_/, '').replace(/_/g, '-');
    }
    get argparseFlag() {
        if (this.fishKind !== 'ARGPARSE')
            return this.name;
        const flagName = this.argparseFlagName;
        if (flagName.length === 1) {
            return `-${flagName}`;
        }
        return `--${flagName}`;
    }
    isArgparseCompletionFlag(node) {
        if (this.fishKind === 'ARGPARSE')
            return false;
        if (node.parent && (0, node_types_1.isCommandWithName)(node, 'complete')) {
            const flagName = this.argparseFlagName;
            if (node.previousSibling) {
                return flagName.length === 1
                    ? options_1.Option.create('-s', '--short').matches(node.previousSibling)
                    : options_1.Option.create('-l', '--long').matches(node.previousSibling);
            }
        }
        return false;
    }
    isCommandCompletionFlag(node) {
        if (this.fishKind === 'COMPLETE')
            return false;
        if (node.parent && (0, node_types_1.isCommandWithName)(node.parent, 'complete')) {
            if (node.previousSibling) {
                return options_1.Option.create('-c', '--command').matches(node.previousSibling);
            }
        }
        return false;
    }
    isEqualLocation(node) {
        if (!node.isNamed || this.focusedNode.equals(node) || !this.nameEqualsNodeText(node)) {
            return false;
        }
        switch (this.fishKind) {
            case 'FUNCTION':
            case 'ALIAS':
                return node.parent && (0, node_types_1.isCommandWithName)(node.parent, 'complete')
                    ? !(0, node_types_1.isVariableDefinitionName)(node) && !(0, node_types_1.isCommand)(node) && this.isCommandCompletionFlag(node)
                    : !(0, node_types_1.isVariableDefinitionName)(node) && !(0, node_types_1.isCommand)(node);
            case 'ARGPARSE':
                // return !isFunctionDefinitionName(node) && isMatchingCompleteOptionIsCommand(node);
                return !(0, node_types_1.isFunctionDefinitionName)(node) || this.isArgparseCompletionFlag(node);
            case 'SET':
            case 'READ':
            case 'FOR':
            case 'VARIABLE':
                return !(0, node_types_1.isFunctionDefinitionName)(node);
        }
    }
    get path() {
        return (0, translation_1.uriToPath)(this.uri);
    }
    get workspacePath() {
        const path = this.path;
        const pathItems = path.split('/');
        let lastItem = pathItems.at(-1);
        if (lastItem === 'config.fish') {
            return pathItems.slice(0, -1).join('/');
        }
        lastItem = pathItems.at(-2);
        if (['functions', 'completions', 'conf.d'].includes(lastItem)) {
            return pathItems.slice(0, -2).join('/');
        }
        return pathItems.slice(0, -1).join('/');
    }
    get scopeTag() {
        return this.scope.scopeTag;
    }
    get scopeNode() {
        return this.scope.scopeNode;
    }
    toString() {
        return JSON.stringify({
            name: this.name,
            kind: this.kind,
            uri: this.uri,
            detail: this.detail,
            range: this.range,
            selectionRange: this.selectionRange,
            scope: this.scope.scopeTag,
            aliasedNames: this.aliasedNames,
            children: this.children.map(child => child.name),
        }, null, 2);
    }
    equals(other) {
        if (this.fishKind === 'ARGPARSE' && other.fishKind === 'ARGPARSE') {
            const equalNames = this.name === other.name || this.aliasedNames.includes(other.name) || other.aliasedNames.includes(this.name);
            return equalNames &&
                this.uri === other.uri &&
                this.node.equals(other.node);
        }
        const equalNames = this.name === other.name
            ? true
            : this.aliasedNames.includes(other.name) || other.aliasedNames.includes(this.name);
        return equalNames &&
            this.kind === other.kind &&
            this.uri === other.uri &&
            this.range.start.line === other.range.start.line &&
            this.range.start.character === other.range.start.character &&
            this.range.end.line === other.range.end.line &&
            this.range.end.character === other.range.end.character &&
            this.selectionRange.start.line === other.selectionRange.start.line &&
            this.selectionRange.start.character === other.selectionRange.start.character &&
            this.selectionRange.end.line === other.selectionRange.end.line &&
            this.selectionRange.end.character === other.selectionRange.end.character &&
            this.fishKind === other.fishKind;
    }
    equalLocations(other) {
        return this.uri === other.uri
            && this.selectionRange.start.line === other.range.start.line
            && this.selectionRange.start.character === other.range.start.character
            && this.selectionRange.end.line === other.range.end.line
            && this.selectionRange.end.character === other.range.end.character;
    }
    toWorkspaceSymbol() {
        return vscode_languageserver_1.WorkspaceSymbol.create(this.name, this.kind, this.uri, this.selectionRange);
    }
    toDocumentSymbol() {
        return vscode_languageserver_1.DocumentSymbol.create(this.name, this.detail, this.kind, this.range, this.selectionRange, this.children.map(child => child.toDocumentSymbol()));
    }
    toLocation() {
        return vscode_languageserver_1.Location.create(this.uri, this.selectionRange);
    }
    isBefore(other) {
        if (this.fishKind === 'FUNCTION' && other.name === 'argv') {
            return this.range.start.line === other.range.start.line
                && this.range.start.character === other.range.start.character;
        }
        if (this.selectionRange.start.line < other.selectionRange.start.line) {
            return true;
        }
        if (this.selectionRange.start.line === other.selectionRange.start.line) {
            return this.selectionRange.start.character < other.selectionRange.start.character
                && this.selectionRange.end.character < other.selectionRange.end.character;
        }
        return false;
    }
    isAfter(other) {
        if (this.name === 'argv' && other.fishKind === 'FUNCTION') {
            return this.selectionRange.start.line === other.selectionRange.start.line
                && this.selectionRange.start.character === other.selectionRange.start.character;
        }
        if (this.selectionRange.start.line > other.selectionRange.start.line) {
            return true;
        }
        if (this.selectionRange.start.line === other.selectionRange.start.line) {
            return this.selectionRange.start.character > other.selectionRange.start.character;
        }
        return false;
    }
    isAfterRange(range) {
        if (this.selectionRange.start.line > range.start.line) {
            return true;
        }
        if (this.selectionRange.start.line === range.start.line) {
            if (this.selectionRange.end.line === range.end.line) {
                return this.selectionRange.start.character > range.start.character
                    && this.selectionRange.end.character <= range.end.character;
            }
            return this.selectionRange.start.character > range.start.character
                && this.selectionRange.end.line <= range.end.line;
        }
        return false;
    }
    toFoldingRange() {
        return {
            startLine: this.range.start.line,
            endLine: this.range.end.line,
            startCharacter: this.range.start.character,
            endCharacter: this.range.end.character,
            collapsedText: this.name,
            kind: vscode_languageserver_1.FoldingRangeKind.Region,
        };
    }
    equalScopes(other) {
        if (this.scope.scopeNode.equals(other.scope.scopeNode) && this.fishKind === other.fishKind) {
            if ([this.scope.scopeTag, other.scope.scopeTag].includes('inherit')) {
                return this.scope.scopeNode.equals(other.scope.scopeNode);
            }
            else if (this.isGlobal() && other.isGlobal()) {
                return true;
            }
            return this.scope.scopeTag === other.scope.scopeTag;
        }
        return false;
    }
    isLocal() {
        return !this.isGlobal();
    }
    isGlobal() {
        return this.scope.scopeTag === 'global' || this.scope.scopeTag === 'universal';
    }
    isRootLevel() {
        return (0, node_types_1.isTopLevelDefinition)(this.node);
    }
    isSymbolImmutable() {
        if (!config_1.config.fish_lsp_modifiable_paths.some(path => this.path.startsWith(path))) {
            return true;
        }
        return false;
    }
    toMarkupContent() {
        return {
            kind: vscode_languageserver_1.MarkupKind.Markdown,
            value: this.detail,
        };
    }
    /**
     * Optionally include the current document's uri to the hover, this will determine
     * if a range is local to the current document (local ranges include hover range)
     */
    toHover(currentUri = '') {
        return {
            contents: this.toMarkupContent(),
            range: currentUri === this.uri ? this.selectionRange : undefined,
        };
    }
    scopeContainsNode(node) {
        return this.scope.containsPosition((0, tree_sitter_1.getRange)(node).start);
    }
    containsNode(node) {
        return this.range.start.line <= node.startPosition.row
            && this.range.end.line >= node.endPosition.row;
    }
    containsPosition(position) {
        return this.selectionRange.start.line === position.line
            && this.selectionRange.start.character <= position.character
            && this.selectionRange.end.character >= position.character;
    }
    //
    // Helpers for checking if the symbol is a fish_lsp_* config variable
    //
    /**
     * Checks if the symbol is a key in the `config` object, which means it changes the
     * configuration of the fish-lsp server.
     */
    isConfigDefinition() {
        if (this.kind !== vscode_languageserver_1.SymbolKind.Variable || this.fishKind !== 'SET') {
            return false;
        }
        return Object.keys(config_1.config).includes(this.name);
    }
    /**
     * Checks if a config variable has the `--erase` option set
     */
    isConfigDefinitionWithErase() {
        if (!this.isConfigDefinition())
            return false;
        const eraseOption = options_1.Option.create('-e', '--erase');
        const definitionNode = this.focusedNode;
        const children = (0, set_1.findSetChildren)(this.node)
            .filter(s => s.startIndex < definitionNode.startIndex);
        return children.some(s => (0, options_1.isMatchingOption)(s, eraseOption));
    }
    /**
     * Finds the value nodes of a config variable definition
     */
    findValueNodes() {
        const valueNodes = [];
        if (!this.isConfigDefinition())
            return valueNodes;
        let node = this.focusedNode.nextNamedSibling;
        while (node) {
            if (!isEmptyString(node))
                valueNodes.push(node);
            node = node.nextNamedSibling;
        }
        return valueNodes;
    }
    /**
     * Converts the value nodes of a config variable definition to shell values
     */
    valuesAsShellValues() {
        return this.findValueNodes().map(node => {
            let text = node.text;
            if ((0, node_types_1.isString)(node))
                text = text.slice(1, -1);
            return file_operations_1.SyncFileHelper.expandEnvVars(text);
        });
    }
}
exports.FishSymbol = FishSymbol;
function getLocalSymbols(symbols) {
    return symbols.filter(symbol => symbol.isLocal());
}
function getGlobalSymbols(symbols) {
    return symbols.filter(symbol => symbol.isGlobal());
}
function isSymbol(symbols, kind) {
    return symbols.filter(symbol => symbol.fishKind === kind);
}
function filterLastPerScopeSymbol(symbols) {
    const flatArray = (0, flatten_1.flattenNested)(...symbols);
    const array = [];
    for (const symbol of symbols) {
        const lastSymbol = flatArray.findLast((s) => {
            return s.name === symbol.name && s.kind === symbol.kind && s.uri === symbol.uri
                && s.equalScopes(symbol);
        });
        if (lastSymbol && lastSymbol.equals(symbol)) {
            array.push(symbol);
        }
    }
    return array;
}
function findLocalLocations(symbol, allSymbols, includeSelf = true) {
    const result = [];
    /*
     * Here we need to handle aliases where there exists a function with the same name
     * (A very weird edge case)
     */
    const matchingNodes = allSymbols.filter(s => s.name === symbol.name && !symbol.equalScopes(s))
        .map(s => symbol.fishKind === 'ALIAS' ? s.node : s.scopeNode);
    for (const node of (0, tree_sitter_1.getChildNodes)(symbol.scopeNode)) {
        /** skip nodes that would be considered a match for another symbol */
        if (matchingNodes.some(n => (0, tree_sitter_1.containsNode)(n, node)))
            continue;
        if (symbol.isEqualLocation(node))
            result.push(node);
    }
    return [
        includeSelf && symbol.name !== 'argv' ? symbol.toLocation() : undefined,
        ...result.map(node => symbol.fishKind === 'ARGPARSE'
            ? vscode_languageserver_1.Location.create(symbol.uri, (0, argparse_1.convertNodeRangeWithPrecedingFlag)(node))
            : vscode_languageserver_1.Location.create(symbol.uri, (0, tree_sitter_1.getRange)(node))),
    ].filter(Boolean);
}
function findMatchingLocations(symbol, allSymbols, document, rootNode) {
    const result = [];
    const matchingNodes = allSymbols.filter(s => s.name === symbol.name && !symbol.equalScopes(s))
        .map(s => symbol.fishKind === 'ALIAS' ? s.node : s.scopeNode);
    for (const node of (0, tree_sitter_1.getChildNodes)(rootNode)) {
        if (matchingNodes.some(n => (0, tree_sitter_1.containsNode)(n, node)))
            continue;
        if (symbol.isEqualLocation(node)) {
            result.push(node);
        }
    }
    return result.map(node => symbol.fishKind === 'ARGPARSE'
        ? vscode_languageserver_1.Location.create(document.uri, (0, argparse_1.convertNodeRangeWithPrecedingFlag)(node))
        : vscode_languageserver_1.Location.create(document.uri, (0, tree_sitter_1.getRange)(node)));
}
function removeLocalSymbols(symbol, symbols) {
    return symbols.filter(s => s.name === symbol.name && !symbol.equalScopes(s) && !s.equals(symbol));
}
function isEmptyString(node) {
    return (0, node_types_1.isString)(node) && node.text.length === 2;
}
/**
 * Formats a tree of FishSymbols into a string with proper indentation
 * @param symbols Array of FishSymbol objects to format
 * @param indentLevel Initial indentation level (optional, defaults to 0)
 * @returns A string representing the formatted tree
 */
function formatFishSymbolTree(symbols, indentLevel = 0) {
    let result = '';
    const indentString = '  '; // 2 spaces per indent level
    for (const symbol of symbols) {
        const indent = indentString.repeat(indentLevel);
        const scopeTag = symbol.scope?.scopeTag || 'unknown';
        result += `${indent}${symbol.name} (${symbol.fishKind}) (${scopeTag})\n`;
        // Recursively format children with increased indent
        if (symbol.children && symbol.children.length > 0) {
            result += formatFishSymbolTree(symbol.children, indentLevel + 1);
        }
    }
    return result;
}
function buildNested(document, node, ...children) {
    const firstNamedChild = node.firstNamedChild;
    const newSymbols = [];
    switch (node.type) {
        case 'function_definition':
            newSymbols.push(...(0, function_1.processFunctionDefinition)(document, node, children));
            break;
        case 'for_statement':
            newSymbols.push(...(0, for_1.processForDefinition)(document, node, children));
            break;
        case 'command':
            if (!firstNamedChild?.text)
                break;
            switch (firstNamedChild.text) {
                case 'set':
                    newSymbols.push(...(0, set_1.processSetCommand)(document, node, children));
                    break;
                case 'read':
                    newSymbols.push(...(0, read_1.processReadCommand)(document, node, children));
                    break;
                case 'argparse':
                    newSymbols.push(...(0, argparse_1.processArgparseCommand)(document, node, children));
                    break;
                case 'alias':
                    newSymbols.push(...(0, alias_1.processAliasCommand)(document, node, children));
                    break;
                case 'export':
                    newSymbols.push(...(0, export_1.processExportCommand)(document, node, children));
                    break;
                default:
                    break;
            }
            break;
    }
    return newSymbols;
}
function processNestedTree(document, ...nodes) {
    const symbols = [];
    /** add argv to script files */
    if (!document.isAutoloadedUri()) {
        const programNode = nodes.find(node => node.type === 'program');
        if (programNode)
            symbols.push(...(0, function_1.processArgvDefinition)(document, programNode));
    }
    for (const node of nodes) {
        // Process children first (bottom-up approach)
        const childSymbols = processNestedTree(document, ...node.children);
        // Process the current node and integrate children
        const newSymbols = buildNested(document, node, ...childSymbols);
        if (newSymbols.length > 0) {
            // If we created symbols for this node, add them (they should contain children)
            symbols.push(...newSymbols);
        }
        else if (childSymbols.length > 0) {
            // If no new symbols from this node but we have child symbols, bubble them up
            symbols.push(...childSymbols);
        }
        // If neither condition is met, we add nothing
    }
    return symbols;
}
