"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getReferences = getReferences;
exports.implementationLocation = implementationLocation;
exports.getArgparseLocations = getArgparseLocations;
exports.getLeadingDashCount = getLeadingDashCount;
const vscode_languageserver_1 = require("vscode-languageserver");
const node_types_1 = require("./utils/node-types");
const tree_sitter_1 = require("./utils/tree-sitter");
const complete_1 = require("./parsing/complete");
const options_1 = require("./parsing/options");
const logger_1 = require("./logger");
const argparse_1 = require("./parsing/argparse");
const Locations = __importStar(require("./utils/locations"));
const translation_1 = require("./utils/translation");
/**
 * get all the references for a symbol, including the symbol's definition
 * @param analyzer the analyzer
 * @param document the document
 * @param position the position of the symbol
 * @param localOnly if true, only return local references inside current document
 * @return the locations of the symbol
 */
function getReferences(analyzer, document, position, localOnly = false) {
    const startTime = performance.now();
    const locations = [];
    const symbol = analyzer.getDefinition(document, position);
    if (!symbol)
        return [];
    if (symbol.fishKind === 'ARGPARSE') {
        locations.push(symbol.toLocation());
        locations.push(...getArgparseLocations(analyzer, symbol, localOnly));
        const endTime = performance.now();
        const duration = ((endTime - startTime) / 1000).toFixed(2); // Convert to seconds with 2 decimal places
        // logging performance
        logger_1.logger.info({
            isArgparse: true,
            document: (0, translation_1.uriToReadablePath)(document.uri),
            position: `${position.line}:${position.character}`,
            symbol: symbol.name,
            locations: locations.length,
            message: `getReferences() took ${duration} ms`,
        });
        return locations;
    }
    // only search for local references if the symbol is local definition
    // NOTICE: we don't do this for argparse locations cause they might not be local
    if (symbol.isLocal())
        localOnly = true;
    locations.push(symbol.toLocation());
    const symbolLocations = findSymbolLocations(analyzer, symbol, localOnly);
    // add unique locations
    for (const location of symbolLocations) {
        if (!locations.some(loc => Locations.Location.equals(loc, location))) {
            locations.push(location);
        }
    }
    const endTime = performance.now();
    const duration = ((endTime - startTime) / 1000).toFixed(2); // Convert to seconds with 2 decimal places
    // logging performance
    logger_1.logger.info({
        isArgparse: false,
        document: (0, translation_1.uriToReadablePath)(document.uri),
        position: `${position.line}:${position.character}`,
        symbol: symbol.name,
        locations: locations.length,
        message: `getReferences() took ${duration} ms`,
    });
    return locations;
}
/**
 * bi-directional jump to either definition or completion definition
 * @param analyzer the analyzer
 * @param document the document
 * @param position the position of the symbol
 * @return the locations of the symbol, should be a lower number of locations than getReferences
 */
function implementationLocation(analyzer, document, position) {
    const locations = [];
    const node = analyzer.nodeAtPoint(document.uri, position.line, position.character);
    if (!node)
        return [];
    const symbol = analyzer.getDefinition(document, position);
    if (!symbol)
        return [];
    const newLocations = getReferences(analyzer, document, position)
        .filter(location => location.uri !== document.uri);
    if (newLocations.some(s => s.uri === symbol.uri)) {
        locations.push(symbol.toLocation());
        return locations;
    }
    if (newLocations.some(s => s.uri.includes('completions/'))) {
        locations.push(newLocations.find(s => s.uri.includes('completions/')));
        return locations;
    }
    locations.push(symbol.toLocation());
    return locations;
}
/**
 * find all the symbol locations for a normal symbol (not an `argparse` flag)
 */
function findSymbolLocations(analyzer, symbol, localOnly = false) {
    const locations = [];
    if (symbol.kind !== vscode_languageserver_1.SymbolKind.Function && symbol.kind !== vscode_languageserver_1.SymbolKind.Variable)
        return [];
    const locationsCallback = (n, doc) => {
        if (localOnly && doc.uri !== symbol.uri)
            return false;
        // check if the node is a local symbol
        if (symbol.isLocal() && doc.uri === symbol.uri) {
            return symbol.scopeContainsNode(n) && n.text === symbol.name && !symbol.focusedNode.equals(n);
        }
        if (symbol.isGlobal()) {
            // get all the local symbols for the current document, and remove any node that is redefined in the local scope
            const localSymbols = analyzer.cache.getFlatDocumentSymbols(doc.uri)
                .filter(s => s.name === symbol.name && s.isLocal());
            if (localSymbols.length > 0 && localSymbols.some(s => s.scopeContainsNode(n))) {
                return false;
            }
            // remove `complete ... -s opt -l opt` entries for variables
            if (symbol.kind === vscode_languageserver_1.SymbolKind.Variable && n.parent) {
                const isCompletion = (0, complete_1.isCompletionCommandDefinition)(n.parent);
                if (isCompletion)
                    return false;
            }
            // remove `complete ... -l cmdname` entries, keep `complete -c cmdname` for functions
            if (symbol.kind === vscode_languageserver_1.SymbolKind.Function && (0, node_types_1.isCompleteCommandName)(n)) {
                return n.text === symbol.name && !symbol.focusedNode.equals(n);
            }
            else if (symbol.kind === vscode_languageserver_1.SymbolKind.Function && n.parent && (0, node_types_1.isCommandWithName)(n.parent, 'complete')) {
                return false;
            }
            // remove non command name entries for functions
            // if (symbol.kind === SymbolKind.Function && n.parent && isCommandWithName(n.parent, 'command', 'type', 'builtin', 'functions')) {
            //   return n.text === symbol.name && !symbol.focusedNode.equals(n);
            // }
            if (symbol.kind === vscode_languageserver_1.SymbolKind.Function && n.parent && !(0, node_types_1.isCommandWithName)(n.parent, symbol.name) && n.parent.firstChild?.equals(n)) {
                return false;
            }
            return n.text === symbol.name && !symbol.focusedNode.equals(n);
        }
        return false;
    };
    const localDocument = analyzer.getDocument(symbol.uri);
    const matchingNodes = [];
    if (localOnly && localDocument) {
        const localNodes = analyzer.getNodes(symbol.uri)
            .filter(n => locationsCallback(n, localDocument));
        matchingNodes.push({ uri: localDocument.uri, nodes: localNodes });
    }
    else {
        matchingNodes.push(...analyzer.findNodes((n, doc) => locationsCallback(n, doc)));
    }
    // create the new locations
    for (const { uri, nodes } of matchingNodes) {
        for (const node of nodes) {
            const range = (0, tree_sitter_1.getRange)(node);
            locations.push(vscode_languageserver_1.Location.create(uri, range));
        }
    }
    return locations;
}
/**
 * Handle finding the locations of an argparse flag, including any completion definition
 * or usage of the flag itself.
 */
function getArgparseLocations(analyzer, symbol, localOnly = false) {
    const result = [];
    const parentName = symbol.parent?.name
        || symbol.scopeNode.firstNamedChild?.text
        || symbol.scopeNode.text;
    const document = analyzer.getDocument(symbol.uri);
    /**
     * Ensure that our document includes all possible global completion location
     */
    if (document) {
        result.push(...(0, argparse_1.getGlobalArgparseLocations)(analyzer, document, symbol));
    }
    const matchingNodes = analyzer.findNodes((n, document) => {
        if (localOnly && document.uri !== symbol.uri)
            return false;
        // complete -c parentName -s ... -l flag-name
        if ((0, argparse_1.isCompletionArgparseFlagWithCommandName)(n, parentName, symbol.argparseFlagName)) {
            return true;
        }
        // parentName --flag-name
        if (n.parent
            && (0, node_types_1.isCommandWithName)(n.parent, parentName)
            && (0, node_types_1.isOption)(n)
            && (0, node_types_1.isMatchingOption)(n, options_1.Option.fromRaw(symbol?.argparseFlag))) {
            return true;
        }
        // _flag_name in scope
        if (document.uri === symbol.uri
            && symbol.scopeContainsNode(n)
            && n.text === symbol.name) {
            return true;
        }
        return false;
    });
    matchingNodes.forEach(({ uri, nodes }) => {
        if (!nodes)
            return;
        nodes.forEach(node => {
            let range = (0, tree_sitter_1.getRange)(node);
            if ((0, node_types_1.isOption)(node)) {
                range = {
                    start: {
                        line: range.start.line,
                        character: range.start.character + getLeadingDashCount(node),
                    },
                    end: {
                        line: range.end.line,
                        character: range.end.character + 1,
                    },
                };
            }
            if (!result.some(loc => Locations.Location.equals(loc, { uri, range }))) {
                result.push(vscode_languageserver_1.Location.create(uri, range));
            }
        });
    });
    return result;
}
function getLeadingDashCount(node) {
    if (!node || !node.text)
        return 0;
    const text = node.text;
    let count = 0;
    for (let i = 0; i < text.length; i++) {
        if (text[i] === '-') {
            count++;
        }
        else {
            break;
        }
    }
    return count;
}
